import collections.abc as collections_abc
import inspect
import sys
import warnings
from abc import ABC, abstractmethod
from dataclasses import dataclass, fields, is_dataclass
from datetime import date, datetime
from enum import Enum, IntEnum
from typing import Any, Dict, Iterable, List, Optional, Tuple, Type, Union

if sys.version_info >= (3, 9):
    from typing import _AnnotatedAlias as AnnotatedAlias

from typing import _GenericAlias as GenericAlias
from typing import get_type_hints
from uuid import UUID

from openapidocs.common import Format
from openapidocs.v3 import (
    APIKeySecurity,
    Components,
    Example,
    Header,
    HTTPSecurity,
    Info,
    MediaType,
    OAuth2Security,
    OpenAPI,
    OpenIdConnectSecurity,
    Operation,
    Parameter,
    ParameterLocation,
    PathItem,
    Reference,
    RequestBody,
)
from openapidocs.v3 import Response as ResponseDoc
from openapidocs.v3 import (
    Schema,
    SecurityRequirement,
    Server,
    Tag,
    ValueFormat,
    ValueType,
)

from shuttleasgi.server.bindings import (
    Binder,
    BodyBinder,
    CookieBinder,
    HeaderBinder,
    QueryBinder,
    RouteBinder,
    empty,
)
from shuttleasgi.server.openapi.docstrings import (
    DocstringInfo,
    get_handler_docstring_info,
)
from shuttleasgi.server.openapi.exceptions import DuplicatedContentTypeDocsException
from shuttleasgi.server.routing import Router

from ..application import Application
from .common import (
    APIDocsHandler,
    ContentInfo,
    EndpointDocs,
    HeaderInfo,
    ParameterInfo,
    ParameterSource,
    RequestBodyInfo,
    ResponseExample,
    ResponseInfo,
    ResponseStatusType,
    response_status_to_str,
)

try:
    from pydantic import BaseModel  # type: ignore
except ImportError:  # pragma: no cover
    # noqa
    BaseModel = ...  # type: ignore


# region PEP 604
try:
    # Python >= 3.10
    from types import UnionType
except ImportError:  # pragma: no cover
    UnionType = ...


def _is_union_type(annotation):
    if UnionType is not ... and isinstance(annotation, UnionType):  # type: ignore
        return True
    return False


# endregion


def get_origin(object_type):
    return getattr(object_type, "__origin__", None)


def check_union(object_type: Any) -> Tuple[bool, Any]:
    if (
        hasattr(object_type, "__origin__")
        and object_type.__origin__ is Union
        or _is_union_type(object_type)
        or (hasattr(object_type, "nullable"))
        or (hasattr(object_type, "is_required"))
    ):
        if hasattr(object_type, "nullable"):
            # Pydantic v1
            return object_type.nullable, object_type
        if hasattr(object_type, "is_required"):
            # Pydantic v2
            return object_type.is_required(), object_type

        if type(None) in object_type.__args__ and len(object_type.__args__) == 2:
            for possible_type in object_type.__args__:
                if type(None) is possible_type:
                    continue
                return True, possible_type
        return type(None) not in object_type.__args__, object_type
    return False, object_type


def is_ignored_parameter(param_name: str, matching_binder: Optional[Binder]) -> bool:
    # if a binder is used, handle only those that can be mapped to a type of
    # OpenAPI Documentation parameter's location
    if matching_binder and not isinstance(
        matching_binder,
        (
            QueryBinder,
            RouteBinder,
            HeaderBinder,
            CookieBinder,
        ),
    ):
        return True

    return param_name == "request" or param_name == "services"


@dataclass
class FieldInfo:
    name: str
    type: Union[Type, Schema]


class ObjectTypeHandler(ABC):
    """
    Abstract class describing the interface used to generate an OpenAPI Documentation
    schema for an object type.
    """

    @abstractmethod
    def handles_type(self, object_type) -> bool:
        """
        Returns a value indicating whether the given type is handled but this
        object type handler.
        """

    @abstractmethod
    def get_type_fields(self, object_type, register_type) -> List[FieldInfo]:
        """
        Returns a set of fields to be used for the handled type.
        """


class DataClassTypeHandler(ObjectTypeHandler):
    def handles_type(self, object_type) -> bool:
        return is_dataclass(object_type)

    def get_type_fields(self, object_type, register_type) -> List[FieldInfo]:
        return [FieldInfo(field.name, field.type) for field in fields(object_type)]


def _try_is_subclass(object_type, check_type):
    try:
        return issubclass(object_type, check_type)
    except TypeError:
        return False


class PydanticModelTypeHandler(ObjectTypeHandler):
    def handles_type(self, object_type) -> bool:
        if isinstance(object_type, GenericAlias):
            # support for Generic BaseModel here is better since it can extract items
            # type, skip this;
            # anyway using Generic with BaseModel doesn't seem to be recommended by
            # pydantic:
            # https://pydantic-docs.helpmanual.io/usage/types/#generic-classes-as-types
            return False
        if (
            BaseModel is not ...
            and inspect.isclass(object_type)
            and _try_is_subclass(object_type, BaseModel)  # type: ignore
        ):
            return True
        return False

    def _allow_none(self, field_info):
        try:
            # Pydantic v1
            return field_info.allow_none
        except AttributeError:
            # Pydantic v2
            return not field_info.is_required()

    def _handle_any_of(self, any_of, types_args, register_type):
        for item in any_of:
            if "$ref" in item:
                obj_type_name = item["$ref"].split("/")[-1]
                corresponding_type = next(
                    obj for obj in types_args if obj.__name__ == obj_type_name
                )
                register_type(corresponding_type)
                item["$ref"] = f"#/components/schemas/{obj_type_name}"
            yield item

    def _open_api_v2_field_schema_to_type(
        self, field_info, schema, register_type=None
    ) -> Union[Type, Schema]:
        if "$ref" in schema:
            if field_info is None:
                return Schema()
            try:
                # Pydantic v1
                return field_info.outer_type_
            except AttributeError:
                # Pydantic v2
                return field_info.annotation

        value_type = schema.get("type")
        nullable = self._allow_none(field_info) if field_info is not None else False

        if value_type == "null":
            # OpenAPI 3.1
            return Schema(type="null")

        if value_type is None:
            if "anyOf" in schema:
                any_of = schema["anyOf"]

                if any("$ref" in item for item in any_of):
                    any_of = list(
                        self._handle_any_of(
                            any_of, field_info.annotation.__args__, register_type
                        )
                    )
                return Schema(
                    any_of=any_of,
                    title=schema.get("title"),
                )

        if value_type == "string":
            return Schema(
                type=ValueType.STRING,
                min_length=schema.get("minLength"),
                max_length=schema.get("maxLength"),
                format=schema.get("format"),
                nullable=nullable,
            )

        if value_type == "boolean":
            return Schema(type=ValueType.BOOLEAN, nullable=nullable)

        if value_type == "file":
            # TODO: improve support for file type,
            # see "Considerations for File Uploads"
            # at https://swagger.io/specification/
            return Schema(type=ValueType.STRING, format=ValueFormat.BINARY)

        if value_type == "number":
            return Schema(
                type=ValueType.NUMBER,
                minimum=schema.get("exclusiveMinimum"),
                maximum=schema.get("exclusiveMaximum"),
                format=ValueFormat.FLOAT,
                nullable=nullable,
            )

        if value_type == "integer":
            return Schema(
                type=ValueType.INTEGER,
                minimum=schema.get("exclusiveMinimum"),
                maximum=schema.get("exclusiveMaximum"),
                format=ValueFormat.INT64,
                nullable=nullable,
            )

        if value_type == "array":
            if field_info is not None:
                return self._get_array_outer_type(field_info)
            else:
                return list

        return Schema()

    def _get_array_outer_type(self, field_info):
        try:
            # Pydantic v1
            # example: typing.List[tests.test_openapi_v3.PydCat]
            return field_info.outer_type_
        except AttributeError:
            # Pydantic v2
            # Here we support only simple types
            return List[field_info.annotation.__args__[0]]

    def _get_fields_info(self, object_type):
        try:
            if hasattr(object_type, "model_fields"):
                # Pydantic v2
                return object_type.model_fields
            else:
                # Pydantic v1
                return object_type.__fields__
        except AttributeError:
            return dict()

    def _get_object_schema(self, object_type):
        if hasattr(object_type, "model_json_schema"):
            # Pydantic v2
            return object_type.model_json_schema()
        else:
            # Pydantic v1
            return object_type.schema()

    def get_type_fields(self, object_type, register_type) -> List[FieldInfo]:
        # to support all pydantic special types, here we rely on its schema,
        # but since we want to be able to generate OpenAPI Documentation V3 (or V2 for
        # that matter), we extract basic information we need; also using the type's
        # __fields__ when available
        schema = self._get_object_schema(object_type)
        properties = schema["properties"]

        fields_info = self._get_fields_info(object_type)

        return [
            FieldInfo(
                name,
                self._open_api_v2_field_schema_to_type(
                    fields_info.get(name, None), value, register_type
                ),
            )
            for name, value in properties.items()
        ]


SecuritySchemes = Union[
    OAuth2Security,
    OpenIdConnectSecurity,
    HTTPSecurity,
    APIKeySecurity,
]


class OpenAPIHandler(APIDocsHandler[OpenAPI]):
    """
    Handles the automatic generation of OpenAPI Documentation, specification v3
    for a web application, exposing methods to enrich the documentation with details
    through decorators.
    """

    def __init__(
        self,
        *,
        info: Info,
        ui_path: str = "/docs",
        json_spec_path: str = "/openapi.json",
        yaml_spec_path: str = "/openapi.yaml",
        preferred_format: Format = Format.JSON,
        anonymous_access: bool = True,
        tags: Optional[List[Tag]] = None,
        security_schemes: Optional[
            Dict[
                str,
                SecuritySchemes,
            ]
        ] = None,
    ) -> None:
        super().__init__(
            ui_path=ui_path,
            json_spec_path=json_spec_path,
            yaml_spec_path=yaml_spec_path,
            preferred_format=preferred_format,
            anonymous_access=anonymous_access,
        )
        self.info = info
        self._tags = tags
        self.components = Components()
        self._objects_references: Dict[Any, Reference] = {}
        self.servers: List[Server] = []
        self.common_responses: Dict[ResponseStatusType, ResponseDoc] = {}
        self._object_types_handlers: List[ObjectTypeHandler] = [
            DataClassTypeHandler(),
            PydanticModelTypeHandler(),
        ]
        self._binder_docs: Dict[Type[Binder], Iterable[Union[Parameter, Reference]]] = (
            {}
        )
        self.security_schemes = security_schemes or {}

    @property
    def object_types_handlers(self) -> List[ObjectTypeHandler]:
        return self._object_types_handlers

    def get_ui_page_title(self) -> str:
        return self.info.title

    def _iter_tags(self, paths: Dict[str, PathItem]) -> Iterable[str]:
        methods = (
            "get",
            "post",
            "put",
            "delete",
            "options",
            "head",
            "patch",
            "trace",
        )

        for path in paths.values():
            for method in methods:
                prop: Optional[Operation] = getattr(path, method)

                if prop is not None and prop.tags is not None:
                    yield from prop.tags

    def _tags_from_paths(self, paths: Dict[str, PathItem]) -> List[Tag]:
        unique_tags = set(self._iter_tags(paths))
        return [Tag(name=name) for name in sorted(unique_tags)]

    def generate_documentation(self, app: Application) -> OpenAPI:
        self._optimize_binders_docs()
        self._generate_security_schemes()
        paths = self.get_paths(app)
        return OpenAPI(
            info=self.info,
            paths=paths,
            components=self.components,
            tags=self._tags or self._tags_from_paths(paths),
        )

    def get_paths(self, app: Application, path_prefix: str = "") -> Dict[str, PathItem]:
        own_paths = self.get_routes_docs(app.router, path_prefix)

        if app.mount_registry.mounted_apps and app.mount_registry.handle_docs:
            for route in app.mount_registry.mounted_apps:
                if isinstance(route.handler, Application):
                    full_prefix = route.pattern.decode().rstrip("/*")
                    if path_prefix:
                        full_prefix = path_prefix.rstrip("/") + full_prefix

                    child_docs = self.get_paths(
                        route.handler,
                        full_prefix,
                    )
                    own_paths.update(child_docs)

        return own_paths

    def get_type_name_for_generic(
        self,
        object_type: GenericAlias,
        context_type_args: Optional[Dict[Any, Type]] = None,
    ) -> str:
        """
        This method returns a type name for a generic type.
        """
        assert isinstance(object_type, GenericAlias), "This method requires a generic"
        # Note: by default returns a string respectful of this requirement:
        # $ref values must be RFC3986-compliant percent-encoded URIs
        # Therefore, a generic that would be expressed in Python: Example[Foo, Bar]
        # and C# or TypeScript Example<Foo, Bar>
        # Becomes here represented as: ExampleOfFooAndBar
        origin = get_origin(object_type)
        args = object_type.__args__
        args_repr = "And".join(
            self.get_type_name(arg, context_type_args) for arg in args
        )
        return f"{self.get_type_name(origin)}Of{args_repr}"

    def get_type_name(
        self, object_type, context_type_args: Optional[Dict[Any, Type]] = None
    ) -> str:
        if context_type_args and object_type in context_type_args:
            object_type = context_type_args.get(object_type)
        if isinstance(object_type, GenericAlias):
            return self.get_type_name_for_generic(object_type, context_type_args)
        if hasattr(object_type, "__name__"):
            return object_type.__name__
        if object_type is Union:
            # Python 3.9 and 3.8 would fall here, Union type has a "_name" property but
            # no "__name__"
            return "Union"
        raise ValueError(
            f"Cannot obtain a name for object_type parameter: {object_type!r}"
        )

    def _generate_security_schemes(self):
        for name, scheme in self.security_schemes.items():
            self.register_security_scheme(name, scheme)

    def register_security_scheme(self, name: str, scheme: SecuritySchemes):
        if self.components.security_schemes is None:
            self.components.security_schemes = {}

        self.components.security_schemes[name] = scheme

    def register_schema_for_type(self, object_type: Type) -> Reference:
        stored_ref = self._get_stored_reference(object_type, None)
        if stored_ref:
            return stored_ref

        schema = self.get_schema_by_type(object_type)
        if isinstance(schema, Schema):
            return self._register_schema(schema, self.get_type_name(object_type))
        return schema

    def _register_schema(self, schema: Schema, name: str) -> Reference:
        if self.components.schemas is None:
            self.components.schemas = {}

        if name in self.components.schemas:
            counter = 0
            base_name = name
            while name in self.components.schemas:
                counter += 1
                name = f"{base_name}{counter}"

        self.components.schemas[name] = schema
        return Reference(f"#/components/schemas/{name}")

    def _can_handle_class_type(self, object_type: Type) -> bool:
        return (
            any(
                handler.handles_type(object_type)
                for handler in self._object_types_handlers
            )
            or object_type in self._types_schemas
        )

    def _handle_type_having_explicit_schema(self, object_type: Type) -> Reference:
        schema = self._types_schemas[object_type]
        type_name = self.get_type_name(object_type, {})
        reference = self._register_schema(
            schema,
            type_name,
        )
        self._objects_references[object_type] = reference
        self._objects_references[type_name] = reference
        return reference

    def _get_schema_for_class(self, object_type: Type) -> Reference:
        assert self._can_handle_class_type(object_type)

        if object_type in self._types_schemas:
            return self._handle_type_having_explicit_schema(object_type)

        required: List[str] = []
        properties: Dict[str, Union[Schema, Reference]] = {}

        for field in self.get_fields(object_type):
            is_optional, child_type = check_union(field.type)
            if not is_optional:
                required.append(field.name)

            if isinstance(child_type, str):
                # this is a forward reference, we need to obtain a full type here
                annotations = get_type_hints(object_type)

                if field.name in annotations:
                    child_type = annotations[field.name]

            properties[field.name] = self.get_schema_by_type(
                child_type, root_optional=is_optional
            )

        return self._handle_object_type(object_type, properties, required)

    def _handle_object_type(
        self,
        object_type: Type,
        properties: Dict[str, Union[Schema, Reference]],
        required: List[str],
        context_type_args: Optional[Dict[Any, Type]] = None,
    ) -> Reference:
        return self._handle_object_type_schema(
            object_type,
            context_type_args,
            Schema(
                type=ValueType.OBJECT,
                required=required or None,
                properties=properties,
            ),
        )

    def _handle_object_type_schema(
        self, object_type, context_type_args, schema: Schema
    ):
        type_name = self.get_type_name(object_type, context_type_args)
        reference = self._register_schema(schema, type_name)
        self._objects_references[object_type] = reference
        self._objects_references[type_name] = reference
        return reference

    def _get_stored_reference(
        self, object_type: Type[Any], type_args: Optional[Dict[Any, Type]] = None
    ) -> Optional[Reference]:
        if object_type in self._objects_references:
            # if object_type is a generic, it can be like
            # Example[~T] while type_args can have the information: {~T: Foo}
            # in such case; check
            return self._objects_references[object_type]

        if type_args:
            type_name = self.get_type_name(object_type, type_args)
            if type_name in self._objects_references:
                return self._objects_references[type_name]

        return None

    def get_schema_by_type(
        self,
        object_type: Union[Type, Schema],
        type_args: Optional[Dict[Any, Type]] = None,
        *,
        root_optional: Optional[bool] = None,
    ) -> Union[Schema, Reference]:
        if isinstance(object_type, Schema):
            return object_type

        stored_ref = self._get_stored_reference(object_type, type_args)
        if stored_ref:
            return stored_ref

        is_optional, child_type = check_union(object_type)
        schema = self._get_schema_by_type(child_type, type_args)
        if isinstance(schema, Schema):
            if is_optional or root_optional:
                schema.nullable = True
            else:
                schema.nullable = False
        return schema

    def _get_schema_by_type(
        self, object_type: Type[Any], type_args: Optional[Dict[Any, Type]] = None
    ) -> Union[Schema, Reference]:
        stored_ref = self._get_stored_reference(object_type, type_args)
        if stored_ref:  # pragma: no cover
            return stored_ref

        if sys.version_info >= (3, 9):
            if isinstance(object_type, AnnotatedAlias):
                # Replace Annotated object type with the original type
                object_type = getattr(object_type, "__origin__")

        if self._can_handle_class_type(object_type):
            return self._get_schema_for_class(object_type)

        schema = self._try_get_schema_for_simple_type(object_type)
        if schema:
            return schema

        # List, Set, Tuple are handled first than GenericAlias
        schema = self._try_get_schema_for_iterable(object_type, type_args)
        if schema:
            return schema

        if isinstance(object_type, GenericAlias):
            schema = self._try_get_schema_for_generic(object_type, type_args)
            if schema:
                return schema

        if inspect.isclass(object_type):
            schema = self._try_get_schema_for_enum(object_type)
        return schema or Schema()

    def _try_get_schema_for_simple_type(self, object_type: Type) -> Optional[Schema]:
        if object_type is str:
            return Schema(type=ValueType.STRING)

        if object_type is int:
            return Schema(type=ValueType.INTEGER, format=ValueFormat.INT64)

        if object_type is float:
            return Schema(type=ValueType.NUMBER, format=ValueFormat.FLOAT)

        if object_type is bool:
            return Schema(type=ValueType.BOOLEAN)

        if object_type is UUID:
            return Schema(type=ValueType.STRING, format=ValueFormat.UUID)

        if object_type is date:
            return Schema(type=ValueType.STRING, format=ValueFormat.DATE)

        if object_type is datetime:
            return Schema(type=ValueType.STRING, format=ValueFormat.DATETIME)

        return None

    def _try_get_schema_for_iterable(
        self, object_type: Type, context_type_args: Optional[Dict[Any, Type]] = None
    ) -> Optional[Schema]:
        if object_type in {list, set, tuple}:
            # the user didn't specify the item type
            return Schema(type=ValueType.ARRAY, items=Schema(type=ValueType.STRING))

        origin = get_origin(object_type)

        if not origin or origin not in {list, set, tuple, collections_abc.Sequence}:
            return None

        # can be List, List[str] or list[str] (Python 3.9),
        # note: it could also be union if it wasn't handled above for dataclasses
        try:
            type_args = object_type.__args__  # type: ignore
        except AttributeError:  # pragma: no cover
            item_type = str
        else:
            item_type = next(iter(type_args), str)

        if context_type_args and item_type in context_type_args:
            item_type = context_type_args.get(item_type)

        return Schema(
            type=ValueType.ARRAY,
            items=self.get_schema_by_type(item_type, context_type_args),
        )

    def get_fields(self, object_type: Any) -> List[FieldInfo]:
        for handler in self._object_types_handlers:
            if handler.handles_type(object_type):
                return handler.get_type_fields(
                    object_type, self.register_schema_for_type
                )

        return []

    def _try_get_schema_for_generic(
        self, object_type: Type, context_type_args: Optional[Dict[Any, Type]] = None
    ) -> Optional[Reference]:
        origin = get_origin(object_type)

        if origin is Union:
            schema = Schema(
                ValueType.OBJECT,
                any_of=[
                    self.get_schema_by_type(child_type, context_type_args)
                    for child_type in object_type.__args__
                ],
            )
            return self._handle_object_type_schema(
                object_type, context_type_args, schema
            )

        required: List[str] = []
        properties: Dict[str, Union[Schema, Reference]] = {}

        args = object_type.__args__
        parameters = origin.__parameters__
        type_args = dict(zip(parameters, args))

        for field in self.get_fields(origin):
            is_optional, child_type = check_union(field.type)
            if not is_optional:
                required.append(field.name)

            if isinstance(child_type, str):
                warnings.warn(
                    f"The return type {object_type!r} contains a forward reference "
                    f"for {field.name}. Forward references in Generic types are not "
                    "supported for automatic generation of OpenAPI Documentation. "
                    "For more information see "
                    "https://www.neoteroi.dev/blacksheep/openapi/",
                    UserWarning,
                )
                return None

            if child_type in type_args:
                # example:
                # class Foo(Generic[T]):
                #    item: T
                child_type = type_args.get(child_type)

            properties[field.name] = self.get_schema_by_type(child_type, type_args)

        return self._handle_object_type(
            object_type, properties, required, context_type_args
        )

    def _try_get_schema_for_enum(self, object_type: Type) -> Optional[Schema]:
        if issubclass(object_type, IntEnum):
            return Schema(type=ValueType.INTEGER, enum=[v.value for v in object_type])
        if issubclass(object_type, Enum):
            return Schema(type=ValueType.STRING, enum=[v.value for v in object_type])
        return None

    def _get_body_binder(self, handler: Any) -> Optional[BodyBinder]:
        return next(
            (binder for binder in handler.binders if isinstance(binder, BodyBinder)),
            None,
        )

    def _get_binder_by_name(self, handler: Any, name: str) -> Optional[Binder]:
        return next(
            (binder for binder in handler.binders if binder.parameter_name == name),
            None,
        )

    def _get_body_binder_content_type(
        self,
        body_binder: BodyBinder,
        body_examples: Optional[Dict[str, Union[Example, Reference]]],
    ) -> Dict[str, MediaType]:
        return {
            possible_type: MediaType(
                schema=self.get_schema_by_type(body_binder.expected_type),
                examples=body_examples,
            )
            for possible_type in body_binder.content_type.split(";")
        }

    def get_request_body(self, handler: Any) -> Union[None, RequestBody, Reference]:
        if not hasattr(handler, "binders"):
            return None
        body_binder = self._get_body_binder(handler)

        if body_binder is None:
            return None

        docs = self.get_handler_docs(handler)
        body_info = docs.request_body if docs else None

        body_examples: Optional[Dict[str, Union[Example, Reference]]] = (
            {key: Example(value=value) for key, value in body_info.examples.items()}
            if body_info and body_info.examples
            else None
        )

        return RequestBody(
            content=self._get_body_binder_content_type(body_binder, body_examples),
            required=body_binder.required,
            description=body_info.description if body_info else None,
        )

    def get_parameter_location_for_binder(
        self, binder: Binder
    ) -> Optional[ParameterLocation]:
        if isinstance(binder, RouteBinder):
            return ParameterLocation.PATH
        if isinstance(binder, QueryBinder):
            return ParameterLocation.QUERY
        if isinstance(binder, CookieBinder):
            return ParameterLocation.COOKIE
        if isinstance(binder, HeaderBinder):
            return ParameterLocation.HEADER
        return None

    def _parameter_source_to_openapi_obj(
        self, value: ParameterSource
    ) -> ParameterLocation:
        return ParameterLocation[value.value.upper()]

    def get_parameters(
        self, handler: Any
    ) -> Optional[List[Union[Parameter, Reference]]]:
        if not hasattr(handler, "binders"):
            return None
        binders: List[Binder] = handler.binders
        parameters: Dict[str, Union[Parameter, Reference]] = {}

        docs = self.get_handler_docs(handler)
        parameters_info = (docs.parameters if docs else None) or dict()

        for binder in binders:
            if binder.__class__ in self._binder_docs:
                self._handle_binder_docs(binder, parameters)
                continue

            location = self.get_parameter_location_for_binder(binder)

            if not location:
                # the binder is used for something that is not a parameter
                # expressed in OpenAPI Docs (e.g. a DI service)
                continue

            if location == ParameterLocation.PATH:
                required = True
            else:
                required = binder.required and binder.default is empty

            # did the user specified information about the parameter?
            param_info = parameters_info.get(binder.parameter_name)

            parameters[binder.parameter_name] = Parameter(
                name=binder.parameter_name,
                in_=location,
                required=required or None,
                schema=self.get_schema_by_type(binder.expected_type),
                description=param_info.description if param_info else "",
                example=param_info.example if param_info else None,
            )

        for key, param_info in parameters_info.items():
            if key not in parameters:
                parameters[key] = Parameter(
                    name=key,
                    in_=self._parameter_source_to_openapi_obj(
                        param_info.source or ParameterSource.QUERY
                    ),
                    required=param_info.required,
                    schema=(
                        self.get_schema_by_type(param_info.value_type)
                        if param_info.value_type
                        else None
                    ),
                    description=param_info.description,
                    example=param_info.example,
                )

        return list(parameters.values())

    def get_handler_security(self, handler: Any) -> Optional[List[SecurityRequirement]]:
        docs = self.get_handler_docs(handler)
        if docs and docs.security:
            return [SecurityRequirement(s.name, s.value) for s in docs.security]

        return None

    def _get_media_type_from_content_doc(self, content_doc: ContentInfo) -> MediaType:
        media_type = MediaType()

        if content_doc.type:
            media_type.schema = self.get_schema_by_type(content_doc.type)

        examples = content_doc.examples
        if examples:
            if len(examples) == 1:
                example_item = examples[0]
                if isinstance(example_item, ResponseExample):
                    media_type.example = self.normalize_example(example_item.value)
                else:
                    media_type.example = self.normalize_example(example_item)
            else:
                examples_doc: Dict[str, Union[Example, Reference]] = {}

                for index, example in enumerate(examples):
                    # support something that is not a ResponseExample,
                    # to offer a more concise API
                    if not isinstance(example, ResponseExample):
                        example = ResponseExample(example)

                    if not example.name:
                        example.name = f"example {index}"
                    examples_doc[example.name] = Example(
                        summary=example.summary,
                        description=example.description,
                        value=self.normalize_example(example.value),
                    )
                media_type.examples = examples_doc

        return media_type

    def _get_content_from_response_info(
        self, response_content: Optional[List[ContentInfo]]
    ) -> Optional[Dict[str, Union[MediaType, Reference]]]:
        if not response_content:
            return None

        oad_content: Dict[str, Union[MediaType, Reference]] = {}

        for content in response_content:
            if content.content_type not in oad_content:
                oad_content[content.content_type] = (
                    self._get_media_type_from_content_doc(content)
                )
            else:
                raise DuplicatedContentTypeDocsException(content.content_type)

        return oad_content

    def _get_headers_from_response_info(
        self, headers: Optional[Dict[str, HeaderInfo]]
    ) -> Optional[Dict[str, Union[Header, Reference]]]:
        if headers is None:
            return None

        return {
            key: Header(value.description, self.get_schema_by_type(value.type))
            for key, value in headers.items()
        }

    def get_responses(self, handler: Any) -> Optional[Dict[str, ResponseDoc]]:
        docs = self.get_handler_docs(handler)
        data = docs.responses if docs else None

        # common responses used by the whole application, like responses sent in case
        # of error
        responses = {
            response_status_to_str(key): value
            for key, value in self.common_responses.items()
        }

        if not data:
            # try to generate automatically from the handler's return type annotations
            return_type = getattr(handler, "return_type", ...)

            if return_type is not ...:
                # automatically set response content for status 200,
                # if the user wants major control, it's necessary to use the decorators
                # responses[response_status_to_str(200)] = return_type
                if data is None:
                    data = {}

                if return_type is None:
                    # the user explicitly marked the request handler as returning None,
                    # document therefore HTTP 204 No Content
                    data["204"] = ResponseInfo("Success response", content=[])
                else:
                    # if the return type is optional, generate automatically
                    # the "404" response
                    is_optional, child_type = check_union(return_type)

                    data["200"] = ResponseInfo(
                        "Success response",
                        content=[ContentInfo(child_type, examples=[])],
                    )

                    if is_optional and self.handle_optional_response_with_404:
                        data["404"] = ResponseInfo("Object not found")
            else:
                return responses

        responses.update(
            {
                response_status_to_str(key): (
                    ResponseDoc(
                        description=value.description,
                        content=self._get_content_from_response_info(value.content),
                        headers=self._get_headers_from_response_info(value.headers),
                    )
                    if isinstance(value, ResponseInfo)
                    else ResponseDoc(description=value)
                )
                for key, value in data.items()
            }
        )
        return responses

    def on_docs_generated(self, docs: OpenAPI) -> None:
        docs.servers = self.servers

        self.events.on_docs_created.fire_sync(docs)

    def _merge_documentation(
        self,
        handler,
        endpoint_docs: EndpointDocs,
        docstring_info: DocstringInfo,
    ) -> None:
        if not endpoint_docs.description and docstring_info.description:
            endpoint_docs.description = docstring_info.description

        if not endpoint_docs.summary and docstring_info.summary:
            endpoint_docs.summary = docstring_info.summary

        for param_name, param_info in docstring_info.parameters.items():
            if endpoint_docs.parameters is None:
                endpoint_docs.parameters = {}

            # did the user specify parameter information explicitly, using @docs?
            matching_parameter = endpoint_docs.parameters.get(param_name)

            if matching_parameter is None:
                matching_binder = self._get_binder_by_name(handler, param_name)

                if isinstance(matching_binder, BodyBinder):
                    if endpoint_docs.request_body is None:
                        endpoint_docs.request_body = RequestBodyInfo(
                            description=param_info.description
                        )
                    elif not endpoint_docs.request_body.description:
                        endpoint_docs.request_body.description = param_info.description
                    continue

                if is_ignored_parameter(param_name, matching_binder):
                    # this must not be documented in OpenAPI Documentation!
                    continue

                assert isinstance(endpoint_docs.parameters, dict)
                endpoint_docs.parameters[param_name] = ParameterInfo(
                    value_type=param_info.value_type,
                    required=param_info.required,
                    description=param_info.description,
                    source=param_info.source or ParameterSource.QUERY,
                )
            else:
                matching_parameter.description = param_info.description

                if (
                    matching_parameter.value_type is None
                    and param_info.value_type is not None
                ):
                    matching_parameter.value_type = param_info.value_type

    def _apply_docstring(self, handler, docs: Optional[EndpointDocs]) -> None:
        if not self.use_docstrings:  # pragma: no cover
            return
        docstring_info = get_handler_docstring_info(handler)

        if docstring_info is not None:
            if docs is None:
                docs = self.get_handler_docs_or_set(handler)
            self._merge_documentation(handler, docs, docstring_info)

    def get_operation_id(self, docs: Optional[EndpointDocs], handler) -> str:
        return handler.__name__

    def get_routes_docs(
        self, router: Router, path_prefix: str = ""
    ) -> Dict[str, PathItem]:
        """Obtains a documentation object from the routes defined in a router."""
        paths_doc: Dict[str, PathItem] = {}
        raw_dict = self.router_to_paths_dict(router, lambda route: route)

        for path, conf in raw_dict.items():
            path_item = PathItem()

            for method, route in conf.items():
                handler = self._get_request_handler(route)
                docs = self.get_handler_docs(handler)
                self._apply_docstring(handler, docs)

                operation = Operation(
                    description=self.get_description(handler),
                    summary=self.get_summary(handler),
                    responses=self.get_responses(handler) or {},
                    operation_id=self.get_operation_id(docs, handler),
                    parameters=self.get_parameters(handler),
                    request_body=self.get_request_body(handler),
                    deprecated=self.is_deprecated(handler),
                    tags=self.get_handler_tags(handler),
                    security=self.get_handler_security(handler),
                )
                if docs and docs.on_created:
                    docs.on_created(self, operation)

                self.events.on_operation_created.fire_sync(operation)
                setattr(path_item, method, operation)

            path_value = path_prefix + path
            if path_value != "/" and path_value.endswith("/"):
                path_value = path_value.rstrip("/")
            paths_doc[path_value] = path_item

        self.events.on_paths_created.fire_sync(paths_doc)
        return paths_doc

    def set_binder_docs(
        self,
        binder_type: Type[Binder],
        params_docs: Iterable[Union[Parameter, Reference]],
    ):
        """
        Configures parameters documentation for a given binder type. A binder can
        read values from one or more input parameters, this is why this method supports
        an iterable of Parameter or Reference objects. In most use cases, it is
        desirable to use a Parameter here. Reference objects are configured
        automatically when the documentation is built.
        """
        self._binder_docs[binder_type] = params_docs

    def _handle_binder_docs(
        self, binder: Binder, parameters: Dict[str, Union[Parameter, Reference]]
    ):
        params_docs = self._binder_docs[binder.__class__]

        for i, param_doc in enumerate(params_docs):
            parameters[f"{binder.__class__.__qualname__}_{i}"] = param_doc

    def _optimize_binders_docs(self):
        """
        Optimizes the documentation for custom binders to use references and
        components.parameters, instead of duplicating parameters documentation in each
        operation where they are used.
        """
        new_dict = {}
        params_docs: Iterable[Union[Parameter, Reference]]

        for key, params_docs in self._binder_docs.items():
            new_docs: List[Reference] = []

            for param in params_docs:
                if isinstance(param, Reference):
                    new_docs.append(param)
                else:
                    if self.components.parameters is None:
                        self.components.parameters = {}

                    self.components.parameters[param.name] = param
                    new_docs.append(Reference(f"#/components/parameters/{param.name}"))

            new_dict[key] = new_docs

        self._binder_docs = new_dict



def extend(obj, cls):
    """
    Applies a mixin to an instance of a class.

    This method is used in those scenarios where opting-in for a feature incurs a
    performance fee, so that said fee is paid only when features are used.
    """
    base_cls = obj.__class__
    base_cls_name = obj.__class__.__name__ + "_" + cls.__name__
    obj.__class__ = type(base_cls_name, (cls, base_cls), {})



import glob
import inspect
import os
from pathlib import Path


def get_parent_file():
    """
    Returns __file__ of the caller's parent module.
    """
    try:
        return inspect.stack()[2][1]
    except IndexError:
        return ""


def import_child_modules(root_path: Path):
    """
    Import automatically all modules defined
    under a certain package path.
    """
    path = str(root_path)
    modules = [
        os.path.basename(f)[:-3]
        for f in glob.glob(path + "/*.py")
        if not os.path.basename(f).startswith("_")
    ]
    stripped_path = os.path.relpath(path).replace("/", ".").replace("\\", ".")
    for module in modules:
        __import__(stripped_path + "." + module)



"""
Common types annotations and functions.
"""

from typing import AnyStr, Dict, Iterable, List, Optional, Tuple, Union

from shuttleasgi.url import URL

KeyValuePair = Tuple[AnyStr, AnyStr]
HeadersType = Union[Dict[AnyStr, AnyStr], Iterable[KeyValuePair]]
ParamsType = Union[Dict[AnyStr, AnyStr], Iterable[KeyValuePair]]
URLType = Union[str, bytes, URL]


def _ensure_header_bytes(value: AnyStr) -> bytes:
    return value if isinstance(value, bytes) else value.encode("ascii")


def _ensure_param_str(value: AnyStr) -> str:
    return value if isinstance(value, str) else value.decode("ascii")


def normalize_headers(
    headers: Optional[HeadersType],
) -> Optional[List[Tuple[bytes, bytes]]]:
    if headers is None:
        return None
    if isinstance(headers, dict):
        return [
            (_ensure_header_bytes(key), _ensure_header_bytes(value))  # type: ignore
            for key, value in headers.items()
        ]
    return [
        (_ensure_header_bytes(key), _ensure_header_bytes(value))
        for key, value in headers
    ]


def normalize_params(params: Optional[ParamsType]) -> Optional[List[Tuple[str, str]]]:
    if params is None:
        return None
    if isinstance(params, dict):
        return [
            (_ensure_param_str(key), _ensure_param_str(value))  # type: ignore
            for key, value in params.items()
        ]
    return [(_ensure_param_str(key), _ensure_param_str(value)) for key, value in params]



from typing import Any, List

from shuttleasgi.server.rendering.abc import ModelHandler, Renderer
from shuttleasgi.server.rendering.models import DefaultModelHandler


def default_renderer() -> Renderer:
    from shuttleasgi.server.rendering.jinja2 import JinjaRenderer

    return JinjaRenderer()


class HTMLSettings:
    def __init__(self):
        self._renderer: Renderer | None = None
        self._model_handlers: List[ModelHandler] = [DefaultModelHandler()]

    def use(self, renderer: Renderer):
        self._renderer = renderer

    @property
    def model_handlers(self) -> List[ModelHandler]:
        return self._model_handlers

    @property
    def renderer(self) -> Renderer:
        if self._renderer is None:
            self._renderer = default_renderer()
        return self._renderer

    def model_to_params(self, model: Any) -> Any:
        for handler in self.model_handlers:
            try:
                return handler.model_to_view_params(model)
            except NotImplementedError:
                continue


html_settings = HTMLSettings()



import pytest

from shuttleasgi.server.controllers import Controller
from shuttleasgi.server.headers.cache import (
    CacheControlMiddleware,
    cache_control,
    write_cache_control_response_header,
)
from shuttleasgi.server.routing import RoutesRegistry
from shuttleasgi.testing.helpers import get_example_scope
from shuttleasgi.testing.messages import MockReceive, MockSend

CACHE_CONTROL_PARAMS_EXPECTED = [
    ({"max_age": 0}, b"max-age=0"),
    ({"max_age": 120}, b"max-age=120"),
    ({"shared_max_age": 604800}, b"s-maxage=604800"),
    ({"no_cache": True}, b"no-cache"),
    ({"no_store": True}, b"no-store"),
    ({"must_understand": True, "no_store": True}, b"no-store, must-understand"),
    ({"private": True}, b"private"),
    ({"public": True}, b"public"),
    ({"no_cache": True, "no_store": True}, b"no-cache, no-store"),
    ({"max_age": 0, "must_revalidate": True}, b"max-age=0, must-revalidate"),
    ({"max_age": 0, "proxy_revalidate": True}, b"max-age=0, proxy-revalidate"),
    ({"no_transform": True}, b"no-transform"),
    (
        {"public": True, "max_age": 604800, "immutable": True},
        b"public, max-age=604800, immutable",
    ),
    (
        {"max_age": 604800, "stale_while_revalidate": 86400},
        b"max-age=604800, stale-while-revalidate=86400",
    ),
    (
        {"max_age": 604800, "stale_if_error": 86400},
        b"max-age=604800, stale-if-error=86400",
    ),
]


async def _assert_scenario(app, expected_header: bytes):
    @app.router.get("/no-cache")
    @cache_control(no_cache=True, no_store=True)
    def example_no():
        return "Example"

    await app.start()
    await app(get_example_scope("GET", "/", []), MockReceive(), MockSend())

    response = app.response
    assert response.status == 200
    cache_control_value = response.headers[b"cache-control"]
    assert len(cache_control_value) == 1
    assert cache_control_value[0] == expected_header

    await app(get_example_scope("GET", "/no-cache", []), MockReceive(), MockSend())

    response = app.response
    assert response.status == 200
    cache_control_value = response.headers[b"cache-control"]
    assert len(cache_control_value) == 1
    assert cache_control_value[0] == b"no-cache, no-store"


def test_write_cache_control_response_header_raises_for_priv_pub():
    with pytest.raises(ValueError):
        write_cache_control_response_header(private=True, public=True)


@pytest.mark.asyncio
@pytest.mark.parametrize("params,expected_header", CACHE_CONTROL_PARAMS_EXPECTED)
async def test_cache_control_decorator(app, params, expected_header):
    @app.router.get("/")
    @cache_control(**params)
    def example():
        return "Example"

    await _assert_scenario(app, expected_header)


@pytest.mark.asyncio
@pytest.mark.parametrize("params,expected_header", CACHE_CONTROL_PARAMS_EXPECTED)
async def test_cache_control_in_controller(app, params, expected_header):
    app.controllers_router = RoutesRegistry()
    get = app.controllers_router.get

    class Home(Controller):
        @get("/")
        @cache_control(**params)
        async def index(self):
            return "Example"

    await _assert_scenario(app, expected_header)


@pytest.mark.asyncio
@pytest.mark.parametrize("params,expected_header", CACHE_CONTROL_PARAMS_EXPECTED)
async def test_cache_control_middleware(app, params, expected_header):
    app.middlewares.append(CacheControlMiddleware(**params))

    @app.router.get("/")
    def example():
        return "Example"

    await _assert_scenario(app, expected_header)



"""
This module provides classes to handle OpenID Connect authentication through integration
with OAuth applications, supporting Authorization Code Grant and Hybrid flows.
"""

import logging
from abc import ABC, abstractmethod
from dataclasses import dataclass
from datetime import datetime
from enum import Enum
from typing import Any, AnyStr, Awaitable, Callable, Dict, Optional, Sequence
from urllib.parse import urlencode

import jwt
from guardpost import AuthenticationHandler, Identity, UnauthorizedError
from guardpost.jwts import InvalidAccessToken, JWTValidator
from itsdangerous import Serializer
from itsdangerous.exc import BadSignature
from jwt import InvalidTokenError

from shuttleasgi.cookies import Cookie
from shuttleasgi.exceptions import BadRequest, Unauthorized
from shuttleasgi.messages import Request, Response, get_absolute_url_to_path
from shuttleasgi.server.application import Application, ApplicationEvent
from shuttleasgi.server.authentication.cookie import CookieAuthentication
from shuttleasgi.server.authentication.jwt import JWTBearerAuthentication
from shuttleasgi.server.authorization import allow_anonymous
from shuttleasgi.server.dataprotection import generate_secret, get_serializer
from shuttleasgi.server.headers.cache import cache_control
from shuttleasgi.server.responses import accepted, bad_request, html, json, ok, redirect
from shuttleasgi.utils import ensure_str
from shuttleasgi.utils.aio import FailedRequestError, HTTPHandler


def get_logger() -> logging.Logger:
    logger = logging.getLogger("shuttleasgi.oidc")
    logger.setLevel(logging.INFO)
    return logger


logger = get_logger()


class OpenIDConfiguration:
    """
    Proxy class for a remote OpenID Connect well-known configuration.
    """

    def __init__(self, data) -> None:
        self._data = data

    @property
    def issuer(self) -> str:
        return self._data["issuer"]

    @property
    def jwks_uri(self) -> str:
        return self._data["jwks_uri"]

    @property
    def authorization_endpoint(self) -> str:
        return self._data["authorization_endpoint"]

    @property
    def token_endpoint(self) -> str:
        return self._data["token_endpoint"]

    @property
    def end_session_endpoint(self) -> Optional[str]:
        return self._data.get("end_session_endpoint")


class OpenIDConnectEvent(ApplicationEvent):
    pass


class OpenIDConnectEvents:
    on_id_token_validated: OpenIDConnectEvent
    on_tokens_received: OpenIDConnectEvent
    on_error: OpenIDConnectEvent

    def __init__(self, context) -> None:
        self.on_tokens_received = OpenIDConnectEvent(context)
        self.on_id_token_validated = OpenIDConnectEvent(context)
        self.on_error = OpenIDConnectEvent(context)


@dataclass
class OpenIDSettings:
    client_id: str
    authority: Optional[str] = None
    audience: Optional[str] = None
    client_secret: Optional[str] = None
    discovery_endpoint: Optional[str] = None
    entry_path: str = "/sign-in"
    logout_path: str = "/sign-out"
    post_logout_redirect_path: str = "/"
    callback_path: str = "/authorization-callback"
    refresh_token_path: str = "/refresh-token"
    response_type: str = "code"
    scope: str = "openid profile email"
    redirect_uri: Optional[str] = None
    scheme_name: str = "OpenIDConnect"
    error_redirect_path: Optional[str] = None
    end_session_endpoint: Optional[str] = None


class OpenIDSettingsError(TypeError):
    def __init__(self, message: str) -> None:
        super().__init__(message)


class MissingClientSecretSettingError(OpenIDSettingsError):
    def __init__(self) -> None:
        super().__init__(
            "Missing application client secret, to use the Authorization Code Grant "
            "flow, it is necessary to configure an application secret."
        )


def _get_desired_path(request: Request) -> str:
    base_path = request.base_path
    path = request.path
    query_part = (
        "" if not request.url.query else ("?" + request.url.query.decode("utf8"))
    )
    return f"{base_path}{path}{query_part}"


class ParametersBuilder:
    def __init__(self, settings: OpenIDSettings) -> None:
        self._settings = settings
        self._serializer = get_serializer(purpose=f"{settings.scheme_name}state")

    @property
    def scope(self) -> str:
        return self._settings.scope

    def get_state(self, request: Request) -> Dict[str, str]:
        desired_path = (
            "/"
            if request.path == self._settings.entry_path
            else _get_desired_path(request)
        )
        return {"orig_path": desired_path, "nonce": generate_secret(8)}

    def read_state(self, state: str) -> Dict[str, str]:
        try:
            return self._serializer.loads(state)
        except BadSignature as signature_error:
            logger.error(
                "Failed to parse the request state (%s), invalid signature.",
                str(signature_error),
            )
            raise BadRequest("Invalid state")

    def get_redirect_url(self, request: Request) -> str:
        return str(get_absolute_url_to_path(request, self._settings.callback_path))

    def build_signin_parameters(self, request: Request):
        if self._settings.client_secret:
            # authorization code grant
            response_type = "code"
        else:
            # hybrid flow, requires implicit flow for id_token to be enabled
            response_type = "id_token"

        state = self.get_state(request)
        parameters = {
            "response_type": response_type,
            "response_mode": "form_post",
            "scope": self.scope,
            "client_id": self._settings.client_id,
            "redirect_uri": self._settings.redirect_uri
            or self.get_redirect_url(request),
            "nonce": state.get("nonce") or generate_secret(8),
        }

        if self._settings.audience:
            # Note: Auth0 and Okta use `audience` parameter when
            # the scope includes custom scopes
            parameters["audience"] = self._settings.audience

        if state:
            parameters["state"] = self._serializer.dumps(state)  # type: ignore

        return parameters

    def _require_secret(self):
        if not self._settings.client_secret:
            raise MissingClientSecretSettingError()

    def build_code_grant_parameters(self, request: Request, code: str):
        self._require_secret()

        return {
            "grant_type": "authorization_code",
            "code": code,
            "scope": self.scope,
            "redirect_uri": self._settings.redirect_uri
            or self.get_redirect_url(request),
            "client_id": self._settings.client_id,
            "client_secret": self._settings.client_secret,
        }

    def build_refresh_token_parameters(self, refresh_token: str):
        self._require_secret()

        return {
            "grant_type": "refresh_token",
            "refresh_token": refresh_token,
            "client_id": self._settings.client_id,
            "client_secret": self._settings.client_secret,
        }


class OpenIDConnectError(Exception):
    def __init__(self, message: str) -> None:
        super().__init__(message)


class OpenIDConnectConfigurationError(OpenIDConnectError):
    """
    Exception thrown when there is an error in the programmer's
    defined configuration.
    """


class OpenIDConnectRequestError(OpenIDConnectError):
    def __init__(
        self,
        request_error: FailedRequestError,
        message: str = "Failed web request to the OpenIDConnect provider.",
    ) -> None:
        super().__init__(message + " " + str(request_error))
        self.request_error = request_error


class OpenIDConnectFailedExchangeError(OpenIDConnectRequestError):
    def __init__(
        self,
        request_error: FailedRequestError,
        message: str = (
            "Failed token exchange web request to the OpenIDConnect provider."
        ),
    ) -> None:
        super().__init__(request_error, message=message)


class TokenResponse:
    __slots__ = ("data",)

    def __init__(self, data) -> None:
        self.data = data

    def __repr__(self) -> str:
        return repr(self.data)

    def __str__(self) -> str:
        return str(self.data)

    @property
    def expires_in(self) -> Optional[int]:
        value = self.data.get("expires_in")
        return int(value) if value else None

    @property
    def expires_on(self) -> Optional[int]:
        value = self.data.get("expires_on")
        return int(value) if value else None

    @property
    def token_type(self) -> Optional[str]:
        return self.data.get("token_type")

    @property
    def id_token(self) -> Optional[str]:
        return self.data.get("id_token")

    @property
    def access_token(self) -> Optional[str]:
        return self.data.get("access_token")

    @property
    def refresh_token(self) -> Optional[str]:
        return self.data.get("refresh_token")


@dataclass
class IDToken:
    """
    Stores information about an id_token that was
    obtained and validated: both the original id_token and
    parsed payload.
    """

    value: str
    data: dict

    def __str__(self) -> str:
        return self.value

    @classmethod
    def from_trusted_token(cls, token):
        return cls(
            value=token, data=jwt.decode(token, options={"verify_signature": False})
        )


class TokensStore(ABC):
    """
    Base abstract class for types that can store and restore access tokens and refresh
    tokens in the context of a web request.
    """

    @abstractmethod
    async def store_tokens(
        self,
        request: Request,
        response: Response,
        access_token: str,
        refresh_token: Optional[str],
    ):
        """
        Applies a strategy to store an access token and an optional refresh token for
        the given request and response.
        """

    @abstractmethod
    async def restore_tokens(self, request: Request):
        """
        Applies a strategy to restore an access token and an optional refresh token for
        the given request.
        """

    async def unset_tokens(self, request: Request):
        """
        Optional method, to unset access tokens upon sign-out.
        """


class OpenIDTokensHandler(AuthenticationHandler):
    """
    Abstract type that can handle tokens for requests and responses for the
    OpenID Connect flow. This class is responsible of communicating tokens
    to clients, and restoring tokens context for following requests.
    Handled tokens can be:
    - id_token(s), access_token(s), refresh_token(s)
    """

    @abstractmethod
    async def get_success_response(
        self,
        request: Request,
        id_token: IDToken,
        token_response: Optional[TokenResponse],
        original_path: str,
    ) -> Response:
        """
        Creates the response after a successful sign-in, used to communicate the
        tokens to the client.
        """

    @abstractmethod
    async def get_refresh_tokens_response(
        self,
        request: Request,
        tokens_response: TokenResponse,
    ) -> Response:
        """
        Creates the response after a successful request to refresh tokens, used to
        communicate new tokens to the client.
        """

    @abstractmethod
    async def get_logout_response(
        self, request: Request, post_logout_redirect_path: str
    ) -> Response:
        """
        Creates the response for a log-out / sign-out event.
        """


class CookiesOpenIDTokensHandler(OpenIDTokensHandler):
    """
    Default authentication handler used in OpenID Connect flows. This class uses a
    cookie to set user context (encrypting data read from validated id_tokens) and
    reads the same cookie to restore user context at each following web request.
    Cookie expiration is configured automatically from id_token `exp` claim.

    IMPORTANT: this class is not able to store all tokens when the flow is configured to
    obtain an id_token, an access_token, and a refresh_token because they would exceed
    the maximum cookie size.
    """

    def __init__(
        self,
        auth_handler: Optional[CookieAuthentication] = None,
        tokens_store: Optional["TokensStore"] = None,
    ) -> None:
        self.auth_handler = (
            auth_handler if auth_handler is not None else CookieAuthentication()
        )
        self.tokens_store = tokens_store
        self._serializer = get_serializer(purpose="oidc-tokens-protection")

    async def get_success_response(
        self,
        request: Request,
        id_token: IDToken,
        token_response: Optional[TokenResponse],
        original_path: str,
    ) -> Response:
        """
        Returns a redirect response that includes Set-Cookie headers
        to configure an encrypted id_token and, optionally, also encrypted
        access_token and refresh_token.
        """
        response = redirect(original_path)
        await self._set_tokens_in_response(request, response, id_token, token_response)
        return response

    async def get_refresh_tokens_response(
        self,
        request: Request,
        tokens_response: TokenResponse,
    ) -> Response:
        response = ok()
        await self._set_tokens_in_response(
            request,
            response,
            IDToken.from_trusted_token(tokens_response.id_token),
            tokens_response,
        )
        return response

    async def _set_tokens_in_response(
        self,
        request: Request,
        response: Response,
        id_token: IDToken,
        token_response: TokenResponse,
    ):
        self.auth_handler.set_cookie(
            id_token.data, response, secure=request.scheme == "https"
        )

        if (
            token_response is not None
            and self.tokens_store
            and token_response.access_token is not None
        ):
            # ability to store access and refresh tokens
            await self.tokens_store.store_tokens(
                request,
                response,
                token_response.access_token,
                token_response.refresh_token,
            )

    async def get_logout_response(
        self, request: Request, post_logout_redirect_path: str
    ) -> Response:
        response = redirect(post_logout_redirect_path)
        self.auth_handler.unset_cookie(response)
        if self.tokens_store is not None:
            await self.tokens_store.unset_tokens(request)
        return response

    async def authenticate(self, context: Request) -> Optional[Identity]:
        await self.auth_handler.authenticate(context)

        if self.tokens_store:
            await self.tokens_store.restore_tokens(context)


class HTMLStorageType(Enum):
    SESSION = "sessionStorage"
    LOCAL = "localStorage"


class JWTOpenIDTokensHandler(OpenIDTokensHandler):
    """
    OpenID Connect authentication handler that uses HTML documents and the HTML5 storage
    API to exchange tokens with the client. This class doesn't use any cookie to
    configure user context, and expects an instance of JWTBearerAuthentication to
    authenticate users using JWT Bearer tokens that the client will include in each web
    request. This class enables more advanced scenarios for authenticating users, with
    the benefit that APIs can be reused more efficiently (as they don't require
    anti-forgery measures anymore) and handle `Authorization: Bearer ...` headers
    instead, but the negative side that automatic redirects are less useful.
    """

    def __init__(
        self,
        auth_handler: JWTBearerAuthentication,
        *,
        id_token_key: str = "ID_TOKEN",
        access_token_key: str = "ACCESS_TOKEN",
        refresh_token_key: str = "REFRESH_TOKEN",
        storage_type: HTMLStorageType = HTMLStorageType.SESSION,
        redirect_timeout: int = 50,
    ) -> None:
        self.auth_handler = auth_handler
        self.id_token_key = id_token_key
        self.access_token_key = access_token_key
        self.refresh_token_key = refresh_token_key
        self.storage_type = storage_type
        self.redirect_timeout = redirect_timeout
        self._serializer = get_serializer(purpose="oidc-tokens-protection")

    async def get_success_response(
        self,
        request: Request,
        id_token: IDToken,
        token_response: Optional[TokenResponse],
        original_path: str,
    ) -> Response:
        """
        Returns a redirect response that includes Set-Cookie headers
        to configure an encrypted id_token and, optionally, also encrypted
        access_token and refresh_token.
        """
        response = html(
            self._get_html(
                original_path,
                id_token.value,
                (
                    token_response.access_token
                    if token_response is not None
                    and token_response.access_token is not None
                    else ""
                ),
                (
                    self.protect_refresh_token(token_response.refresh_token)
                    if token_response is not None
                    and token_response.refresh_token is not None
                    else ""
                ),
            )
        )

        return response

    async def get_refresh_tokens_response(
        self,
        request: Request,
        token_response: TokenResponse,
    ) -> Response:
        refresh_token = token_response.data.get("refresh_token")
        if refresh_token:
            token_response.data["refresh_token"] = self.protect_refresh_token(
                refresh_token
            )
        return json(token_response.data)

    async def get_logout_response(
        self, request: Request, post_logout_redirect_path: str
    ) -> Response:
        return html(self._get_html(post_logout_redirect_path, "", "", ""))

    def _get_storage_code(self) -> str:
        return f"window.{self.storage_type.value}"

    def _get_html(
        self, redirect_path: str, id_token: str, access_token: str, refresh_token: str
    ) -> str:
        storage_code = self._get_storage_code()
        return f"""
<!DOCTYPE html>
<html>
  <head>
    <title>OIDC.</title>
    <meta charset="utf-8" />
    <meta content='text/html; charset=utf-8' http-equiv='Content-Type'>
  </head>
  <body>
    <script>
    (function () {{
        {storage_code}.setItem("{self.id_token_key}", "{id_token}");
        {storage_code}.setItem("{self.access_token_key}", "{access_token}");
        {storage_code}.setItem("{self.refresh_token_key}", "{refresh_token}");
        setTimeout(function () {{
            window.location.replace("{redirect_path}")
        }}, {str(self.redirect_timeout)})
    }})();
    </script>
  </body>
</html>
        """.strip()

    def _get_refresh_token_header_name(self) -> bytes:
        return b"X-" + self.refresh_token_key.replace("_", "-").encode()

    def protect_refresh_token(self, refresh_token: str) -> str:
        return self._serializer.dumps(refresh_token)  # type: ignore

    def restore_refresh_token(self, context: Request) -> None:
        refresh_token_header = context.get_first_header(
            self._get_refresh_token_header_name()
        )

        if refresh_token_header:
            try:
                value = self._serializer.loads(refresh_token_header)
            except BadSignature:
                logger.info(
                    "Discarding refresh token (%s), invalid signature.",
                    self.refresh_token_key,
                )
            else:
                if context.user is None:
                    context.user = Identity({})
                context.user.refresh_token = value

    async def authenticate(self, context: Request) -> Optional[Identity]:
        await self.auth_handler.authenticate(context)

        self.restore_refresh_token(context)


class TokenType(Enum):
    ID_TOKEN = "id_token"
    ACCESS_TOKEN = "access_token"
    REFRESH_TOKEN = "refresh_token"


class CookiesTokensStore(TokensStore):
    """
    A class that can store access and refresh tokens in encrypted form in cookies.

    Beware that cookies size can be problematic when storing all information in cookies.
    If this is the case, consider implementing a type of `BaseTokensStore` that uses
    other ways to store and restore tokens (e.g. Redis Cache, etc.).
    """

    def __init__(
        self,
        scheme_name: str = "OpenIDConnect",
        secret_keys: Optional[Sequence[str]] = None,
        serializer: Optional[Serializer] = None,
        refresh_token_path: str = "/refresh-token",
    ) -> None:
        self._scheme_name: str
        self._access_token_name: str
        self._refresh_token_name: str
        self._refresh_token_path = refresh_token_path
        self.scheme_name = scheme_name

        self.serializer = serializer or get_serializer(
            secret_keys, f"{scheme_name}tokens"
        )

    @property
    def scheme_name(self) -> str:
        return self._scheme_name

    @scheme_name.setter
    def scheme_name(self, value: str) -> None:
        self._scheme_name = value
        self._access_token_cookie_name = f"{self._scheme_name}.at".lower()
        self._refresh_token_cookie_name = f"{self._scheme_name}.rt".lower()

    async def store_tokens(
        self,
        request: Request,
        response: Response,
        access_token: str,
        refresh_token: Optional[str],
    ) -> None:
        secure = request.scheme == "https"
        self.set_cookie(
            response,
            self._access_token_cookie_name,
            access_token,
            secure=secure,
            expires=None,
        )

        len_combined = len(access_token) + len(access_token)
        logger.info(f"COOKIES LEN: {len_combined}")

        if refresh_token:
            self.set_cookie(
                response,
                self._refresh_token_cookie_name,
                refresh_token,
                secure=secure,
                expires=None,
                path=self._refresh_token_path,
            )

    async def restore_tokens(self, request: Request):
        await self._restore_token(
            request, self._access_token_cookie_name, TokenType.ACCESS_TOKEN
        )
        await self._restore_token(
            request, self._refresh_token_cookie_name, TokenType.REFRESH_TOKEN
        )

    def set_cookie(
        self,
        response: Response,
        cookie_name: str,
        data: AnyStr,
        secure: bool,
        expires: Optional[datetime] = None,
        path: str = "/",
    ) -> None:
        value = self.serializer.dumps(data)  # type: ignore
        response.set_cookie(
            Cookie(
                cookie_name,
                ensure_str(value),  # type: ignore
                domain=None,
                path=path,
                http_only=True,
                secure=secure,
                expires=expires,
            )
        )

    async def _restore_token(
        self, request: Request, cookie_name: str, token_type: TokenType
    ) -> None:
        cookie = request.get_cookie(cookie_name)

        if cookie is None:
            pass
        else:
            try:
                value = self.serializer.loads(cookie)
            except BadSignature:
                logger.info(
                    "Discarding token (%s), invalid signature.",
                    cookie_name,
                )
            else:
                if request.user:
                    if token_type == TokenType.ACCESS_TOKEN:
                        request.user.access_token = value
                    elif token_type == TokenType.REFRESH_TOKEN:
                        request.user.refresh_token = value
        return None


class OpenIDConnectHandler:
    def __init__(
        self,
        settings: OpenIDSettings,
        auth_handler: OpenIDTokensHandler,
        parameters_builder: Optional[ParametersBuilder] = None,
    ) -> None:
        self._settings = settings
        self._configuration: Optional[OpenIDConfiguration] = None
        self._http_handler: HTTPHandler = HTTPHandler()
        self.events = OpenIDConnectEvents(self)
        self.parameters_builder = parameters_builder or ParametersBuilder(settings)
        self._jwt_validator: Optional[JWTValidator] = None
        self._auth_handler = auth_handler

    @property
    def auth_handler(self) -> OpenIDTokensHandler:
        return self._auth_handler

    @property
    def settings(self) -> OpenIDSettings:
        return self._settings

    async def get_jwt_validator(self) -> JWTValidator:
        if self._jwt_validator is None:
            configuration = await self.get_openid_configuration()

            self._jwt_validator = JWTValidator(
                require_kid=True,
                keys_url=configuration.jwks_uri,
                valid_issuers=[configuration.issuer],
                valid_audiences=[self._settings.client_id],
            )
        return self._jwt_validator

    def get_well_known_openid_configuration_url(self) -> str:
        if self._settings.discovery_endpoint:
            return self._settings.discovery_endpoint

        if not self._settings.authority:
            raise OpenIDConnectConfigurationError(
                "Missing `authority` or `discovery_endpoint` in OpenIDSettings. "
                "To fix, define one of the two."
            )

        return (
            self._settings.authority.rstrip("/") + "/.well-known/openid-configuration"
        )

    async def fetch_openid_configuration(self) -> OpenIDConfiguration:
        try:
            data = await self._http_handler.fetch_json(
                self.get_well_known_openid_configuration_url()
            )
        except FailedRequestError as request_error:
            logger.error(
                "Failed to fetch OpenID Connect configuration from the remote endpoint."
                "Inspect the exception details for more details on the cause of the "
                "failure.",
                exc_info=request_error,
            )
            raise OpenIDConnectRequestError(request_error)
        else:
            logger.debug(
                "Fetched OpenID Connect configuration from the remote endpoint."
            )
            return OpenIDConfiguration(data)

    async def get_openid_configuration(self) -> OpenIDConfiguration:
        if self._configuration is None:
            self._configuration = await self.fetch_openid_configuration()

        return self._configuration

    async def redirect_to_sign_in(self, request: Request) -> Response:
        redirect_url = await self.get_redirect_uri(request)
        return redirect(redirect_url)

    async def get_redirect_uri(self, request: Request) -> str:
        openid_conf = await self.get_openid_configuration()
        parameters = self.parameters_builder.build_signin_parameters(request)
        authorization_endpoint = openid_conf.authorization_endpoint
        return authorization_endpoint + "?" + urlencode(parameters)

    async def handle_error(self, request: Request, data: Dict[str, Any]) -> Response:
        """
        Handles an error received from the identity provider.
        """
        query = "?" + urlencode({"error": data.get("error", "unknown")})
        if self._settings.error_redirect_path:
            return redirect(self._settings.error_redirect_path + query)
        return redirect("/" + query)

    async def handle_auth_redirect(self, request: Request) -> Response:
        """
        Handles the redirect after the user interacted with the sign-in page of a remote
        authorization server.
        """
        data = await request.form()

        if data is None or not data:
            return accepted()

        error = data.get("error")

        if error:
            logger.error(
                "Received a post request with error message to the OIDC authorization "
                "callback endpoint: %s",
                error,
            )
            await self.events.on_error.fire(data)
            return await self.handle_error(request, data)

        # the following code validates the state and restores the original path
        # the user was trying to access before being redirected to the OIDC sign-in
        state = data.get("state")
        if isinstance(state, str):
            state = self.parameters_builder.read_state(state)
            redirect_path = state.get("orig_path", "/")
        else:
            redirect_path = "/"

        settings = self._settings

        id_token = data.get("id_token")
        token_response = None

        if id_token and not settings.client_secret:
            logger.debug("Successfully obtained an id_token for a user.")
        else:
            code = data.get("code")

            if settings.client_secret and isinstance(code, str):
                # extra call to fetch an access token
                token_response = await self.exchange_token(request, code)

                await self.events.on_tokens_received.fire(token_response)

                if not id_token and token_response.id_token:
                    id_token = token_response.id_token
            else:
                raise BadRequest("Expected either an error, an id_token, or a code.")

        if not isinstance(id_token, str):
            # This can happen legitimately if OIDC settings are configured to retrieve
            # only an access token. For example, when using Auth0 and configuring a
            # single scope that does not include openid or profile. In such cases, it is
            # unclear what should be done, especially since access tokens are not stored
            # by default. The user of the library might still being handling the token
            # response using dedicated event, or using a token_store.
            parsed_id_token = {}
        else:
            parsed_id_token = await self.validate_id_token(id_token)

            if isinstance(state, dict) and state.get("nonce") != parsed_id_token.get(
                "nonce"
            ):
                raise OpenIDConnectError("nonce mismatch error")

            await self.events.on_id_token_validated.fire(parsed_id_token)

        if not isinstance(id_token, str):
            id_token = ""
        return await self._auth_handler.get_success_response(
            request, IDToken(id_token, parsed_id_token), token_response, redirect_path
        )

    async def exchange_token(self, request: Request, code: str) -> TokenResponse:
        configuration = await self.get_openid_configuration()

        code_grant_parameters = self.parameters_builder.build_code_grant_parameters(
            request, code
        )

        try:
            data = await self._http_handler.post_form(
                configuration.token_endpoint, code_grant_parameters
            )
        except FailedRequestError as request_error:
            logger.error(
                "Failed to exchange an authorization code with an access token. "
                "Inspect the exception details for more details on the cause of the "
                "failure.",
                exc_info=request_error,
            )
            raise OpenIDConnectFailedExchangeError(request_error)
        else:
            logger.debug("Exchanged a code with id_token and access_token for a user.")
            return TokenResponse(data)

    async def handle_refresh_token_request(self, request: Request) -> Response:
        if request.user is None:
            return bad_request("Missing user context.")

        refresh_token = request.user.refresh_token

        if not refresh_token:
            return bad_request("Missing refresh_token in request context.")

        token_response = await self.refresh_token(refresh_token)

        await self.events.on_tokens_received.fire(token_response)

        return await self._auth_handler.get_refresh_tokens_response(
            request, token_response
        )

    async def refresh_token(self, refresh_token: str) -> TokenResponse:
        configuration = await self.get_openid_configuration()

        refresh_token_parameters = (
            self.parameters_builder.build_refresh_token_parameters(refresh_token)
        )

        try:
            data = await self._http_handler.post_form(
                configuration.token_endpoint, refresh_token_parameters
            )
        except FailedRequestError as request_error:
            logger.error(
                "Failed to exchange a refresh token with an access token. "
                "Inspect the exception details for more details on the cause of the "
                "failure.",
                exc_info=request_error,
            )
            raise OpenIDConnectFailedExchangeError(request_error)
        else:
            logger.debug(
                "Exchanged a refresh token with a fresh access_token for a user."
            )
            return TokenResponse(data)

    async def validate_id_token(self, raw_id_token: str) -> Any:
        jwt_validator = await self.get_jwt_validator()
        try:
            return await jwt_validator.validate_jwt(raw_id_token)
        except (InvalidAccessToken, InvalidTokenError) as ex:
            logger.error("Invalid id_token: %s", str(ex))
            raise Unauthorized(f"Invalid id_token: {ex}")

    async def handle_logout_redirect(self, request: Request) -> Response:
        # TODO: obtain the logout redirect from OIDC.end_session_endpoint
        # and make what is necessary to logout from the remote identity provider.
        # Note that there are differences among providers:
        # e.g. Okta requires the original id_token

        # Auth0 does not provide end_session_endpoint in the discovery endpoint
        # AAD is the simplest scenario because it's a simple redirect to the
        # end_session_endpoint.
        return await self._auth_handler.get_logout_response(
            request, self._settings.post_logout_redirect_path
        )


class ChallengeMiddleware:
    def __init__(
        self, request_handler: Callable[[Request], Awaitable[Response]]
    ) -> None:
        self.request_handler = request_handler

    async def __call__(self, request: Request, handler):
        try:
            return await handler(request)
        except (Unauthorized, UnauthorizedError):
            if (
                request.user is None
                or not request.user.is_authenticated()
                and request.method in {"GET", "HEAD"}
            ):
                return await self.request_handler(request)
            raise


def use_openid_connect(
    app: Application,
    settings: OpenIDSettings,
    auth_handler: Optional[OpenIDTokensHandler] = None,
    parameters_builder: Optional[ParametersBuilder] = None,
    is_default: bool = True,
) -> OpenIDConnectHandler:
    """
    Configures an application to use OpenID Connect, integrating with an identity
    provider such as Auth0, Okta, Azure Active Directory.

    Parameters
    ----------
    app : Application
        The application to be configured to handle OpenID Connect.
    settings : OpenIDSettings
        Basic OAuth settings, and other settings to handle the OIDC flow.
    auth_handler : Optional[OpenIDTokensHandler]
        Lets specify the object that is responsible of handling responses to communicate
        tokens to the client and restoring user context for web requests.
        If not specified, the default CookiesOpenIDTokensHandler class is used.
        ShuttleASGI offers two kinds of this class out of the box: one that uses cookies
        and, for more advanced situations, one that uses the HTML5 storage API to set
        tokens on the client side. See [the OIDC
        examples](https://github.com/Neoteroi/BlackSheep-Examples/tree/main/oidc)
        for more information.
    parameters_builder : Optional[ParametersBuilder]
        ParametersBuilder used to build parameters for OAuth requests, by default None
    is_default : bool, optional
        If true, the application is configured to automatically redirect
        not-authenticated users to the sign-in endpoint, by default True

    Returns
    -------
    OpenIDConnectHandler
        Instance of a class that handles the OIDC integration.
    """
    scheme_name = settings.scheme_name or "OpenIDConnect"

    if auth_handler is None:
        auth_handler = CookiesOpenIDTokensHandler(
            CookieAuthentication(
                cookie_name=scheme_name.lower(), auth_scheme=scheme_name
            ),
        )

    app.use_authentication().add(auth_handler)

    handler = OpenIDConnectHandler(
        settings,
        parameters_builder=parameters_builder,
        auth_handler=auth_handler,
    )

    @allow_anonymous()
    @app.router.get(settings.entry_path)
    @cache_control(no_cache=True, no_store=True)
    async def redirect_to_sign_in(request: Request):
        return await handler.redirect_to_sign_in(request)

    @allow_anonymous()
    @app.router.post(settings.callback_path)
    @cache_control(no_cache=True, no_store=True)
    async def handle_auth_redirect(request: Request):
        return await handler.handle_auth_redirect(request)

    @app.router.get(settings.logout_path)
    @cache_control(no_cache=True, no_store=True)
    async def redirect_to_logout(request: Request):
        return await handler.handle_logout_redirect(request)

    @allow_anonymous()
    @app.router.post(settings.refresh_token_path)
    @cache_control(no_cache=True, no_store=True)
    async def refresh_token(request: Request):
        return await handler.handle_refresh_token_request(request)

    if is_default:

        @app.on_middlewares_configuration
        def insert_challenge_middleware(app):
            app.middlewares.insert(0, ChallengeMiddleware(handler.redirect_to_sign_in))

    return handler



# cython: language_level=3
# Copyright (C) 2018-present Roberto Prevato
#
# This module is part of ShuttleASGI and is released under
# the MIT License https://opensource.org/licenses/MIT


cdef class URL:

    cdef readonly bytes value
    cdef readonly bytes schema
    cdef readonly bytes host
    cdef readonly int port
    cdef readonly bytes path
    cdef readonly bytes query
    cdef readonly bytes fragment
    cdef readonly bint is_absolute

    cpdef URL join(self, URL other)
    cpdef URL base_url(self)
    cpdef URL with_host(self, bytes host)
    cpdef URL with_scheme(self, bytes schema)
    cpdef URL with_query(self, bytes query)


cpdef URL build_absolute_url(
    bytes scheme,
    bytes host,
    bytes base_path,
    bytes path
)



import sys
from datetime import MINYEAR, datetime, timezone

UTC = timezone.utc

MIN_DATETIME = datetime(MINYEAR, 1, 1, tzinfo=None)


def utcnow() -> datetime:
    if sys.version_info < (3, 12):
        return datetime.utcnow()
    return datetime.now(UTC).replace(tzinfo=None)



"""
This module offers methods to return file paths for resources. Its original purpose is
to provide contents of static files stored for conveniency in the shuttleasgi.server.res
package.
"""

try:
    from importlib.resources import files

    def get_resource_file_path(anchor, file_name: str) -> str:
        return str(files(anchor) / file_name)

except ImportError:
    # Python 3.8
    import pkg_resources

    def get_resource_file_path(anchor, file_name: str) -> str:
        return pkg_resources.resource_filename(anchor, file_name)


def get_resource_file_content(file_name: str) -> str:
    with open(
        get_resource_file_path("shuttleasgi.server.res", file_name),
        mode="rt",
    ) as source:
        return source.read()



from abc import ABC, abstractmethod
from dataclasses import dataclass
from typing import Callable, Optional

from shuttleasgi.messages import Request, Response
from shuttleasgi.server.files.static import get_response_for_static_content
from shuttleasgi.server.resources import get_resource_file_content
from shuttleasgi.utils.time import utcnow

SWAGGER_UI_JS_URL = (
    "https://cdn.jsdelivr.net/npm/swagger-ui-dist@5/swagger-ui-bundle.js"
)
SWAGGER_UI_CSS_URL = "https://cdn.jsdelivr.net/npm/swagger-ui-dist@5/swagger-ui.css"
SWAGGER_UI_FONT = None

REDOC_UI_JS_URL = "https://cdn.jsdelivr.net/npm/redoc@next/bundles/redoc.standalone.js"
REDOC_UI_CSS_URL = None
REDOC_UI_FONT_URL = (
    "https://fonts.googleapis.com/css?family=Montserrat:300,400,700|Roboto:300,400,700"
)


@dataclass
class UIFilesOptions:
    js_url: str
    css_url: Optional[str] = None
    fonts_url: Optional[str] = None


@dataclass
class UIOptions:
    spec_url: str
    page_title: str


class UIProvider(ABC):
    ui_files: UIFilesOptions
    ui_path: str

    def __init__(
        self,
        ui_path: str,
        ui_files: Optional[UIFilesOptions] = None,
    ) -> None:
        super().__init__()
        self.ui_path = ui_path
        self.ui_files = ui_files if ui_files else self.default_ui_files

    @abstractmethod
    def build_ui(self, options: UIOptions) -> None:
        """
        Prepares the UI that will be served by the UI route.
        """

    @abstractmethod
    def get_ui_handler(self) -> Callable[[Request], Response]:
        """
        Returns a request handler for the route that serves a UI.
        """

    @property
    def default_ui_files(self) -> UIFilesOptions: ...


class SwaggerUIProvider(UIProvider):
    def __init__(
        self,
        ui_path: str = "/docs",
        ui_files_options: Optional[UIFilesOptions] = None,
    ) -> None:
        super().__init__(ui_path, ui_files_options)

        self._ui_html: bytes = b""

    def get_openapi_ui_html(self, options: UIOptions) -> str:
        """
        Returns the HTML response to serve the Swagger UI.
        """
        return (
            get_resource_file_content("swagger-ui.html")
            .replace("##SPEC_URL##", options.spec_url)
            .replace("##PAGE_TITLE##", options.page_title)
            .replace("##JS_URL##", self.ui_files.js_url)
            .replace("##CSS_URL##", self.ui_files.css_url or "")
        )

    def build_ui(self, options: UIOptions) -> None:
        self._ui_html = self.get_openapi_ui_html(options).encode("utf8")

    def get_ui_handler(self) -> Callable[[Request], Response]:
        current_time = utcnow().timestamp()

        def get_open_api_ui(request: Request) -> Response:
            return get_response_for_static_content(
                request, b"text/html; charset=utf-8", self._ui_html, current_time
            )

        return get_open_api_ui

    @property
    def default_ui_files(self) -> UIFilesOptions:
        return UIFilesOptions(SWAGGER_UI_JS_URL, SWAGGER_UI_CSS_URL, SWAGGER_UI_FONT)


class ReDocUIProvider(UIProvider):
    def __init__(
        self, ui_path: str = "/redocs", ui_files: Optional[UIFilesOptions] = None
    ) -> None:
        super().__init__(ui_path, ui_files)

        self._ui_html: bytes = b""

    def get_openapi_ui_html(self, options: UIOptions) -> str:
        """
        Returns the HTML response to serve the Swagger UI.
        """
        return (
            get_resource_file_content("redoc-ui.html")
            .replace("##SPEC_URL##", options.spec_url)
            .replace("##PAGE_TITLE##", options.page_title)
            .replace("##JS_URL##", self.ui_files.js_url)
            .replace("##FONT_URL##", self.ui_files.fonts_url or "")
        )

    def build_ui(self, options: UIOptions) -> None:
        self._ui_html = self.get_openapi_ui_html(options).encode("utf8")

    def get_ui_handler(self) -> Callable[[Request], Response]:
        current_time = utcnow().timestamp()

        def get_open_api_ui(request: Request) -> Response:
            return get_response_for_static_content(
                request, b"text/html; charset=utf-8", self._ui_html, current_time
            )

        return get_open_api_ui

    @property
    def default_ui_files(self) -> UIFilesOptions:
        return UIFilesOptions(REDOC_UI_JS_URL, REDOC_UI_CSS_URL, REDOC_UI_FONT_URL)



import json
from typing import Any, Dict, Optional

import jwt
import pytest
from guardpost import AuthorizationContext, Identity, Policy, UnauthorizedError
from guardpost.common import AuthenticatedRequirement
from guardpost.jwks import JWKS, InMemoryKeysProvider, KeysProvider
from pytest import raises
from rodi import Container

from shuttleasgi.messages import Request
from shuttleasgi.server.application import Application
from shuttleasgi.server.authentication import (
    AuthenticateChallenge,
    AuthenticationHandler,
)
from shuttleasgi.server.authentication.jwt import JWTBearerAuthentication
from shuttleasgi.server.authorization import (
    AuthorizationWithoutAuthenticationError,
    Requirement,
    allow_anonymous,
    auth,
    get_www_authenticated_header_from_generic_unauthorized_error,
)
from shuttleasgi.server.di import di_scope_middleware, register_http_context
from shuttleasgi.server.resources import get_resource_file_path
from shuttleasgi.testing.helpers import get_example_scope
from shuttleasgi.testing.messages import MockReceive, MockSend
from tests.test_files_serving import get_folder_path
from tests.utils.application import FakeApplication


def get_file_path(file_name, folder_name: str = "res") -> str:
    return get_resource_file_path("tests", f"{folder_name}/{file_name}")


# region JWTBearer


def test_jwt_bearer_authentication_sets_authority_as_issuer():
    authority = "https://sts.windows.net/a2884dee-52e8-4034-8ce2-6b48e18d1ae7/"
    auth = JWTBearerAuthentication(
        valid_audiences=["api://0ed1cebe-b7ca-45c5-a4bf-a8d586c18d31"],
        authority=authority,
    )
    assert auth._validator._valid_issuers == [authority]


def test_jwt_bearer_authentication_throws_for_missing_issuer():
    with pytest.raises(TypeError):
        JWTBearerAuthentication(valid_audiences=["foo"])


def get_test_jwks() -> JWKS:
    with open(get_file_path("jwks.json"), mode="rt", encoding="utf8") as jwks_file:
        jwks_dict = json.loads(jwks_file.read())
    return JWKS.from_dict(jwks_dict)


@pytest.fixture(scope="session")
def default_keys_provider() -> KeysProvider:
    return InMemoryKeysProvider(get_test_jwks())


def get_token(kid: str, payload: Dict[str, Any], *, fake_kid: Optional[str] = None):
    with open(get_file_path(f"{kid}.pem"), "r") as key_file:
        private_key = key_file.read()

    return jwt.encode(
        payload,
        private_key,
        algorithm="RS256",
        headers={"kid": kid if not fake_kid else fake_kid},
    )


# endregion


class MockAuthHandler(AuthenticationHandler):
    def __init__(self, identity=None):
        if identity is None:
            identity = Identity({"id": "001", "name": "Charlie Brown"}, "JWT")
        self.user = identity

    async def authenticate(self, context: Any) -> Optional[Identity]:
        context.user = self.user
        return context.user


class MockNotAuthHandler(AuthenticationHandler):
    async def authenticate(self, context: Any) -> Optional[Identity]:
        context.user = Identity({"id": "007"})
        # NB: an identity without authentication scheme is treated
        # as anonymous identity
        return context.user


class AccessTokenCrashingHandler(AuthenticationHandler):
    async def authenticate(self, context: Any) -> Optional[Identity]:
        raise AuthenticateChallenge(
            "Bearer",
            None,
            {
                "error": "Invalid access token",
                "error_description": "Access token expired",
            },
        )


class AdminRequirement(Requirement):
    def handle(self, context: AuthorizationContext):
        identity = context.identity

        if identity is not None and identity.claims.get("role") == "admin":
            context.succeed(self)


class AdminsPolicy(Policy):
    def __init__(self):
        super().__init__("admin", AdminRequirement())


@pytest.mark.asyncio
async def test_authentication_sets_identity_in_request(app):
    app.use_authentication().add(MockAuthHandler())

    identity = None

    @app.router.get("/")
    async def home(request):
        nonlocal identity
        identity = request.user
        return None

    app.prepare()
    await app(get_example_scope("GET", "/"), MockReceive(), MockSend())

    assert app.response.status == 204

    assert identity is not None
    assert identity["id"] == "001"
    assert identity["name"] == "Charlie Brown"


@pytest.mark.asyncio
async def test_authorization_forbidden_error_1(app):
    app.use_authentication().add(MockAuthHandler())

    app.use_authorization().add(AdminsPolicy())

    @auth("admin")
    @app.router.get("/")
    async def home():
        return None

    app.prepare()
    await app(get_example_scope("GET", "/"), MockReceive(), MockSend())

    assert app.response.status == 403


@pytest.mark.asyncio
async def test_authorization_policy_success(app):
    admin = Identity({"id": "001", "name": "Charlie Brown", "role": "admin"}, "JWT")

    app.use_authentication().add(MockAuthHandler(admin))

    app.use_authorization().add(AdminsPolicy())

    @auth("admin")
    @app.router.get("/")
    async def home():
        return None

    app.prepare()
    await app(get_example_scope("GET", "/"), MockReceive(), MockSend())

    assert app.response.status == 204


@pytest.mark.asyncio
async def test_authorization_forbidden_error_2(app):
    admin = Identity({"id": "001", "name": "Charlie Brown", "role": "user"}, "JWT")

    app.use_authentication().add(MockAuthHandler(admin))

    app.use_authorization().add(AdminsPolicy())

    @auth("admin")
    @app.router.get("/")
    async def home():
        return None

    app.prepare()
    await app(get_example_scope("GET", "/"), MockReceive(), MockSend())

    assert app.response.status == 403


@pytest.mark.asyncio
async def test_authorization_default_allows_anonymous(app):
    app.use_authentication().add(MockAuthHandler())

    app.use_authorization().add(AdminsPolicy())

    @app.router.get("/")
    async def home():
        return None

    app.prepare()
    await app(get_example_scope("GET", "/"), MockReceive(), MockSend())

    assert app.response.status == 204


@pytest.mark.asyncio
async def test_authorization_supports_default_require_authenticated(app):
    app.use_authentication().add(MockNotAuthHandler())

    app.use_authorization().add(
        AdminsPolicy()
    ).default_policy += AuthenticatedRequirement()

    @app.router.get("/")
    async def home():
        return None

    app.prepare()
    await app(get_example_scope("GET", "/"), MockReceive(), MockSend())

    assert app.response.status == 401


@pytest.mark.asyncio
async def test_static_files_allow_anonymous_by_default(app):
    app.use_authentication().add(MockNotAuthHandler())

    app.use_authorization().add(
        AdminsPolicy()
    ).default_policy += AuthenticatedRequirement()

    @app.router.get("/")
    async def home():
        return None

    app.serve_files(get_folder_path("files"))

    await app.start()

    await app(get_example_scope("GET", "/"), MockReceive(), MockSend())

    assert app.response.status == 401

    await app(get_example_scope("GET", "/lorem-ipsum.txt"), MockReceive(), MockSend())

    assert app.response.status == 200
    content = await app.response.text()
    assert content == "Lorem ipsum dolor sit amet\n"


@pytest.mark.asyncio
async def test_static_files_support_authentication(app):
    app.use_authentication().add(MockNotAuthHandler())

    app.use_authorization().add(
        AdminsPolicy()
    ).default_policy += AuthenticatedRequirement()

    @app.router.get("/")
    async def home():
        return None

    app.serve_files(get_folder_path("files"), allow_anonymous=False)

    await app.start()

    await app(get_example_scope("GET", "/"), MockReceive(), MockSend())

    assert app.response.status == 401

    await app(get_example_scope("GET", "/lorem-ipsum.txt"), MockReceive(), MockSend())

    assert app.response.status == 401


@pytest.mark.asyncio
async def test_static_files_support_authentication_by_route(app):
    app.use_authentication().add(MockNotAuthHandler())

    app.use_authorization().add(
        AdminsPolicy()
    ).default_policy += AuthenticatedRequirement()

    @app.router.get("/")
    async def home():
        return None

    app.serve_files(get_folder_path("files"), allow_anonymous=False)
    app.serve_files(get_folder_path("files2"), allow_anonymous=True, root_path="/login")

    await app.start()

    await app(get_example_scope("GET", "/"), MockReceive(), MockSend())

    assert app.response.status == 401

    await app(get_example_scope("GET", "/lorem-ipsum.txt"), MockReceive(), MockSend())

    assert app.response.status == 401

    await app(get_example_scope("GET", "/login/index.html"), MockReceive(), MockSend())

    assert app.response.status == 200
    content = await app.response.text()
    assert (
        content
        == """<!DOCTYPE html>
<html>
  <head>
    <title>Example.</title>
    <link rel="stylesheet" type="text/css" href="/styles/main.css" />
  </head>
  <body>
    <h1>Lorem ipsum</h1>
    <p>Dolor sit amet.</p>
    <script src="/scripts/main.js"></script>
  </body>
</html>
"""
    )


@pytest.mark.asyncio
async def test_authorization_supports_allow_anonymous(app):
    from shuttleasgi.server.responses import text

    app.use_authentication().add(MockNotAuthHandler())

    app.use_authorization().add(
        AdminsPolicy()
    ).default_policy += AuthenticatedRequirement()

    @allow_anonymous()
    @app.router.get("/")
    async def home():
        return text("Hi There!")

    app.prepare()
    await app(get_example_scope("GET", "/"), MockReceive(), MockSend())

    assert app.response.status == 200


@pytest.mark.asyncio
async def test_authentication_challenge_response(app):
    app.use_authentication().add(AccessTokenCrashingHandler())

    app.use_authorization().add(
        AdminsPolicy()
    ).default_policy += AuthenticatedRequirement()

    @app.router.get("/")
    async def home():
        return None

    app.prepare()
    await app(get_example_scope("GET", "/"), MockReceive(), MockSend())

    assert app.response.status == 401
    header = app.response.get_single_header(b"WWW-Authenticate")

    assert header is not None
    assert header == (
        b'Bearer, error="Invalid access token", '
        b'error_description="Access token expired"'
    )


@pytest.mark.asyncio
async def test_authorization_strategy_without_authentication_raises(app):
    app.use_authorization()

    with raises(AuthorizationWithoutAuthenticationError):
        app.prepare()


@pytest.mark.parametrize(
    "scheme,realm,parameters,expected_value",
    [
        ["Basic", None, None, b"Basic"],
        ["Bearer", "Mushrooms Kingdom", None, b'Bearer realm="Mushrooms Kingdom"'],
        [
            "Bearer",
            "Mushrooms Kingdom",
            {
                "title": "Something",
                "error": "Invalid access token",
                "error_description": "access token expired",
            },
            b'Bearer realm="Mushrooms Kingdom", '
            b'title="Something", '
            b'error="Invalid access token", '
            b'error_description="access token expired"',
        ],
    ],
)
def test_authentication_challenge_error(scheme, realm, parameters, expected_value):
    error = AuthenticateChallenge(scheme, realm, parameters)

    header = error.get_header()
    assert header[0] == b"WWW-Authenticate"
    assert header[1] == expected_value


@pytest.mark.parametrize(
    "exc,expected_value",
    [
        (UnauthorizedError(None, [], scheme="AAD"), b"AAD"),
        (UnauthorizedError(None, [], scheme="example"), b"example"),
        (
            UnauthorizedError(None, [], scheme="Something-Something"),
            b"Something-Something",
        ),
    ],
)
def test_get_www_authenticated_header_from_generic_unauthorized_error(
    exc, expected_value
):
    header = get_www_authenticated_header_from_generic_unauthorized_error(exc)

    assert header is not None
    name, value = header
    assert name == b"WWW-Authenticate"
    assert value == expected_value


@pytest.mark.asyncio
async def test_authorization_default_requires_authenticated_user(app):
    app.use_authentication().add(MockNotAuthHandler())

    app.use_authorization()

    @app.router.get("/")
    async def home():
        return None

    @auth()
    @app.router.get("/admin")
    async def admin():
        return None

    app.prepare()
    await app(get_example_scope("GET", "/"), MockReceive(), MockSend())
    assert app.response.status == 204

    await app(get_example_scope("GET", "/admin"), MockReceive(), MockSend())
    assert app.response.status == 401


@pytest.mark.asyncio
async def test_jwt_bearer_authentication(app, default_keys_provider):
    app.use_authentication().add(
        JWTBearerAuthentication(
            valid_audiences=["a"],
            valid_issuers=["b"],
            keys_provider=default_keys_provider,
        )
    )

    identity: Optional[Identity] = None

    @app.router.get("/")
    async def home(request):
        nonlocal identity
        identity = request.user
        return None

    await app.start()
    await app(get_example_scope("GET", "/"), MockReceive(), MockSend())

    assert app.response is not None
    assert app.response.status == 204

    assert identity is not None
    assert identity.is_authenticated() is False

    # request with valid Bearer Token
    access_token = get_token(
        "0",
        {
            "aud": "a",
            "iss": "b",
            "id": "001",
            "name": "Charlie Brown",
        },
    )
    await app(
        get_example_scope(
            "GET",
            "/",
            extra_headers=[(b"Authorization", b"Bearer " + access_token.encode())],
        ),
        MockReceive(),
        MockSend(),
    )

    assert app.response is not None
    assert app.response.status == 204

    assert identity is not None
    assert identity["id"] == "001"
    assert identity["name"] == "Charlie Brown"

    # request with invalid Bearer Token (invalid audience)
    access_token = get_token(
        "0",
        {
            "aud": "NO",
            "iss": "b",
            "id": "001",
            "name": "Charlie Brown",
        },
    )
    await app(
        get_example_scope(
            "GET",
            "/",
            extra_headers=[(b"Authorization", b"Bearer " + access_token.encode())],
        ),
        MockReceive(),
        MockSend(),
    )

    assert app.response is not None
    assert app.response.status == 204

    assert identity is not None
    assert identity.is_authenticated() is False

    # request with invalid Bearer Token (invalid header)
    access_token = get_token(
        "0",
        {
            "aud": "a",
            "iss": "b",
            "id": "001",
            "name": "Charlie Brown",
        },
    )
    await app(
        get_example_scope(
            "GET",
            "/",
            extra_headers=[(b"Authorization", access_token.encode())],
        ),
        MockReceive(),
        MockSend(),
    )

    assert app.response is not None
    assert app.response.status == 204

    assert identity is not None
    assert identity.is_authenticated() is False


def test_set_authentication_strategy_more_than_once(app: Application):
    auth_strategy = app.use_authentication()
    assert app.use_authentication() is auth_strategy


def test_set_authorization_strategy_more_than_once(app: Application):
    auth_strategy = app.use_authorization()
    assert app.use_authorization() is auth_strategy


class Foo:
    pass


class TestHandler(AuthenticationHandler):
    foo: Foo

    def authenticate(self, context: Request) -> Optional[Identity]:
        context.foo = self.foo  # type: ignore
        return Identity({"test": True})


class TestHandlerReqDep(AuthenticationHandler):
    """Example class to test injection of the Request object. Not recommended."""

    request: Request

    def authenticate(self, context: Request) -> Optional[Identity]:
        assert context is self.request
        return Identity()


@pytest.mark.asyncio
async def test_di_works_with_auth_handlers(app: Application):
    app.services.register(Foo)
    app.services.register(TestHandler)

    auth_strategy = app.use_authentication()
    auth_strategy += TestHandler

    identity: Optional[Identity] = None

    @app.router.get("/")
    async def home(request):
        nonlocal identity
        identity = request.user
        return None

    await app.start()
    await app(get_example_scope("GET", "/"), MockReceive(), MockSend())

    assert isinstance(app, FakeApplication)
    assert app.response is not None
    assert app.response.status == 204


@pytest.mark.asyncio
async def test_di_supports_scoped_auth_handlers(app: Application):
    """
    Verifies that it is possible to have scoped services across request handlers and
    authentication handlers.
    This requires opting-in, adding an additional middleware.
    """

    @app.on_middlewares_configuration
    def enable_scoped_services(_):
        app.middlewares.insert(0, di_scope_middleware)

    assert isinstance(app.services, Container)
    app.services.add_scoped(Foo)
    app.services.register(TestHandler)

    auth_strategy = app.use_authentication()
    auth_strategy += TestHandler

    first_foo: Optional[Foo] = None

    @app.router.get("/")
    async def home(request, foo: Foo):
        nonlocal first_foo
        assert request.foo is foo
        if first_foo is None:
            first_foo = foo
        else:
            # in a second call, a scoped service must be different because the scope
            # is bound to a web request
            assert first_foo is not foo
        return None

    await app.start()
    await app(get_example_scope("GET", "/"), MockReceive(), MockSend())

    assert isinstance(app, FakeApplication)
    assert app.response is not None
    assert app.response.status == 204

    await app(get_example_scope("GET", "/"), MockReceive(), MockSend())


@pytest.mark.asyncio
async def test_di_supports_scoped_auth_handlers_with_request_dep(app: Application):
    """
    Verifies that an authentication handler having Request as dependency, is created
    with the request object.
    """

    register_http_context(app)

    app.services.register(TestHandlerReqDep)

    assert isinstance(app.services, Container)

    auth_strategy = app.use_authentication()
    auth_strategy += TestHandlerReqDep

    @app.router.get("/")
    async def home(request):
        return None

    await app.start()
    await app(get_example_scope("GET", "/"), MockReceive(), MockSend())

    assert isinstance(app, FakeApplication)
    assert app.response is not None
    assert app.response.status == 204

    await app(get_example_scope("GET", "/"), MockReceive(), MockSend())



import inspect
import logging
import re
from abc import ABC, abstractmethod
from collections import defaultdict
from functools import lru_cache
from typing import Any, AnyStr, Callable, Dict, List, Optional, Sequence, Set, Union
from urllib.parse import unquote

from shuttleasgi.common import extend
from shuttleasgi.common.types import (
    HeadersType,
    ParamsType,
    normalize_headers,
    normalize_params,
)
from shuttleasgi.messages import Request
from shuttleasgi.utils import ensure_bytes, ensure_str


class RouteMethod:
    GET = "GET"
    GET_WS = "GET_WS"
    HEAD = "HEAD"
    POST = "POST"
    PUT = "PUT"
    DELETE = "DELETE"
    TRACE = "TRACE"
    OPTIONS = "OPTIONS"
    CONNECT = "CONNECT"
    PATCH = "PATCH"


# for backward compatibility
HTTPMethod = RouteMethod


_route_all_rx = re.compile(b"\\*")
_route_param_rx = re.compile(b"/:([^/]+)")
_mustache_route_param_rx = re.compile(b"/{([^}]+)}")
_angle_bracket_route_param_rx = re.compile(b"/<([^>]+)>")
_named_group_rx = re.compile(b"\\?P<([^>]+)>")
_escaped_chars = {b".", b"[", b"]", b"(", b")"}


class RouteException(Exception):
    """Base class for routing exceptions."""


class RouteDuplicate(RouteException):
    def __init__(self, method, pattern, current_handler):
        method = ensure_str(method)
        pattern = ensure_str(pattern)
        super().__init__(
            f"Cannot register route pattern `{pattern}` for "
            f"`{method}` more than once. "
            f"This pattern is already registered for handler "
            f"{current_handler.__qualname__}."
        )
        self.method = method
        self.pattern = pattern
        self.current_handler = current_handler


class InvalidRouterConfigurationError(RouteException):
    """Base class for router configuration errors"""


class OrphanDefaultRouterError(InvalidRouterConfigurationError):
    """
    Error raised when the default router was configured with routes, but it is not
    associated to any application.
    """

    def __init__(self) -> None:
        super().__init__(
            "Invalid router configuration: the default router was used to register "
            "routes, but it is not associated to any application object. To resolve, "
            "ensure that the router bound to your application is used to register "
            "routes. Do not use routing methods imported from the library."
        )


class SharedRouterError(InvalidRouterConfigurationError):
    """
    Error raised when the more than one application is using the same router.
    Each application object should use a different router.
    """

    def __init__(self) -> None:
        super().__init__(
            "Invalid routers configuration: the same router is used in more "
            "than one Application object. When working with multiple applications, "
            "ensure that each application is configured to use different routers. "
            "For more information, refer to: "
            "https://www.neoteroi.dev/blacksheep/routing/"
        )


class InvalidValuePatternName(RouteException):
    def __init__(self, parameter_pattern_name: str, matched_parameter: str) -> None:
        super().__init__(
            f"Invalid value pattern: {parameter_pattern_name} "
            f"for route parameter {matched_parameter}."
            f"Define a value pattern in the `Route.value_patterns` class "
            f"attribute to configure additional patterns for route values."
        )

        self.parameter_pattern_name = parameter_pattern_name
        self.matched_parameter = matched_parameter


class RouteMatch:
    __slots__ = ("_values", "pattern", "handler")

    def __init__(self, route: "Route", values: Optional[Dict[str, bytes]]):
        self.handler = route.handler
        self.pattern = route.pattern
        self._values: Optional[Dict[str, str]] = (
            {k: unquote(v.decode("utf8")) for k, v in values.items()}
            if values
            else None
        )

    @property
    def values(self) -> Optional[Dict[str, str]]:
        return self._values


class RouteFilter(ABC):
    @abstractmethod
    def handle(self, request: Request) -> bool:
        """
        Returns a value indicating whether a request should be handled by a request
        handler.
        For example, to filter requests by headers.
        """


class HeadersFilter(RouteFilter):
    """
    Filters requests by required headers values.
    """

    def __init__(self, required_headers: HeadersType) -> None:
        self.required_headers = normalize_headers(required_headers) or []

    def handle(self, request: Request) -> bool:
        for key, value in self.required_headers:
            if request.get_first_header(key) != value:
                return False
        return True


class ParamsFilter(RouteFilter):
    """
    Filters requests by required query parameters.
    """

    def __init__(self, required_params: ParamsType) -> None:
        self.required_params = normalize_params(required_params) or []

    def handle(self, request: Request) -> bool:
        for key, value in self.required_params:
            query = request.query.get(key)
            if query is not None and len(query) == 1 and query[0] == value:
                continue
            if query != value:
                return False
        return True


class HostFilter(RouteFilter):
    """
    Filters requests by host value. The comparison is always case insensitive.
    By default the port number is ignored, if present in request headers, unless the
    given host value itself includes the character ":".
    """

    def __init__(self, value: str, ignore_port: bool = True) -> None:
        self._host = value.lower()
        self._ignore_port = ignore_port if ":" not in value else False

    @property
    def host(self) -> str:
        return self._host

    def handle(self, request: Request) -> bool:
        req_host = request.host.lower()
        if self._ignore_port and ":" in req_host:
            req_host = req_host[0 : req_host.index(":")]
        return req_host == self._host


def normalize_filters(
    host: Optional[str] = None,
    headers: Optional[HeadersType] = None,
    params: Optional[ParamsType] = None,
    filters: Optional[List[RouteFilter]] = None,
) -> List[RouteFilter]:
    if filters:
        filters = filters.copy()
    else:
        filters = []

    if headers:
        filters.insert(0, HeadersFilter(headers))

    if params:
        filters.insert(0, ParamsFilter(params))

    if host:
        filters.insert(0, HostFilter(host))

    return filters


def _get_parameter_pattern_fragment(
    parameter_name: bytes, value_pattern: bytes = rb"[^\/]+"
) -> bytes:
    return b"/(?P<" + parameter_name + b">" + value_pattern + b")"


class Route:
    __slots__ = (
        "handler",
        "pattern",
        "param_names",
        "_rx",
    )

    pattern: bytes

    value_patterns = {
        "string": r"[^\/]+",
        "str": r"[^\/]+",
        "path": r".*",
        "int": r"\d+",
        "float": r"\d+(?:\.\d+)?",
        "uuid": r"[a-zA-Z0-9]{8}-[a-zA-Z0-9]{4}-[a-zA-Z0-9]"
        + r"{4}-[a-zA-Z0-9]{4}-[a-zA-Z0-9]{12}",
    }

    def __init__(
        self,
        pattern: Union[str, bytes],
        handler: Any,
    ):
        raw_pattern = self.normalize_pattern(pattern)
        self.handler = handler
        self.pattern = raw_pattern
        rx, param_names = self._get_regex_for_pattern(raw_pattern)
        self._rx = rx
        self.param_names = [name.decode("utf8") for name in param_names]

    @property
    def rx(self) -> re.Pattern:
        return self._rx

    @property
    def has_params(self) -> bool:
        return self._rx.groups > 0

    def _get_regex_for_pattern(self, pattern: bytes):
        """
        Converts a raw pattern into a compiled regular expression that can be used
        to match bytes URL paths, extracting route parameters.
        """
        # TODO: should shuttleasgi support ":" in routes (using escape chars)?
        for c in _escaped_chars:
            if c in pattern:
                pattern = pattern.replace(c, b"\\" + c)

        if b"*" in pattern:
            # throw exception if a star appears more than once
            if pattern.count(b"*") > 1:
                raise RouteException(
                    "A route pattern cannot contain more than one star sign *. "
                    "Multiple star signs are not supported."
                )

            if b"/*" in pattern:
                pattern = _route_all_rx.sub(rb"?(?P<tail>.*)", pattern)
            else:
                pattern = _route_all_rx.sub(rb"(?P<tail>.*)", pattern)

        # support for < > patterns, e.g. /api/cats/<cat_id>
        # but also: /api/cats/<int:cat_id> or /api/cats/<uuid:cat_id> for more
        # granular control on the generated pattern
        if b"<" in pattern:
            pattern = _angle_bracket_route_param_rx.sub(
                self._handle_rich_parameter, pattern
            )

        # support for mustache patterns, e.g. /api/cats/{cat_id}
        # but also: /api/cats/{int:cat_id} or /api/cats/{uuid:cat_id} for more
        # granular control on the generated pattern
        if b"{" in pattern:
            pattern = _mustache_route_param_rx.sub(self._handle_rich_parameter, pattern)

        # route parameters defined using /:name syntax
        if b"/:" in pattern:
            pattern = _route_param_rx.sub(rb"/(?P<\1>[^\/]+)", pattern)

        # NB: following code is just to throw user friendly errors;
        # regex would fail anyway, but with a more complex message
        # 'sre_constants.error: redefinition of group name'
        # we only return param names as they are useful for other things
        param_names = []
        for p in _named_group_rx.finditer(pattern):
            param_name = p.group(1)
            if param_name in param_names:
                raise ValueError(
                    f"cannot have multiple parameters with name: " f"{param_name}"
                )

            param_names.append(param_name)

        if len(pattern) > 1 and not pattern.endswith(b"*"):
            # NB: the /? at the end ensures that a route is matched both with
            # a trailing slash or not
            pattern = pattern + b"/?"
        return re.compile(b"^" + pattern + b"$", re.IGNORECASE), param_names

    def _handle_rich_parameter(self, match: re.Match):
        """
        Handles a route parameter that can include details about the pattern,
        for example:

        /api/cats/<int:cat_id>
        /api/cats/<uuid:cat_id>

        /api/cats/{int:cat_id}
        /api/cats/{uuid:cat_id}
        """
        assert (
            len(match.groups()) == 1
        ), "The regex using this function must handle a single group at a time."

        matched_parameter = next(iter(match.groups()))
        assert isinstance(matched_parameter, bytes)

        if b":" in matched_parameter:
            assert matched_parameter.count(b":") == 1

            raw_pattern_name, parameter_name = matched_parameter.split(b":")
            parameter_pattern_name = raw_pattern_name.decode()
            parameter_pattern = Route.value_patterns.get(parameter_pattern_name)

            if not parameter_pattern:
                raise InvalidValuePatternName(
                    parameter_pattern_name,
                    matched_parameter.decode("utf8"),
                )

            return _get_parameter_pattern_fragment(
                parameter_name, parameter_pattern.encode()
            )
        return _get_parameter_pattern_fragment(matched_parameter)

    def normalize_pattern(self, pattern: Union[str, bytes]) -> bytes:
        if isinstance(pattern, str):
            raw_pattern = pattern.encode("utf8")
        else:
            raw_pattern = pattern

        if raw_pattern == b"":
            raw_pattern = b"/"
        if len(raw_pattern) > 1 and raw_pattern.endswith(b"/"):
            raw_pattern = raw_pattern.rstrip(b"/")

        return raw_pattern

    def __repr__(self) -> str:
        return f"<{self.__class__.__name__} \"{self.pattern.decode('utf8')}\">"

    @staticmethod
    def _normalize_rich_parameter(match: re.Match):
        matched_parameter = next(iter(match.groups()))
        parts = matched_parameter.split(b":")
        parameter_name = parts[1] if len(parts) > 1 else matched_parameter
        return b"/{" + parameter_name + b"}"

    @property
    def mustache_pattern(self) -> str:
        pattern = self.pattern
        if b"<" in pattern:
            pattern = _angle_bracket_route_param_rx.sub(
                self._normalize_rich_parameter, pattern
            )
        if b"{" in pattern:
            pattern = _mustache_route_param_rx.sub(
                self._normalize_rich_parameter, pattern
            )
        return _route_param_rx.sub(rb"/{\1}", pattern).decode("utf8")

    @property
    def full_pattern(self) -> bytes:
        return self._rx.pattern

    def match(self, request: Request) -> Optional[RouteMatch]:
        return self.match_by_path(ensure_bytes(request._path))

    def match_by_path(self, path: bytes) -> Optional[RouteMatch]:
        """
        Returns a match by path - this method can be used only when the route does not
        define any filter.
        """
        if not self.has_params and path.lower() == self.pattern:
            return RouteMatch(self, None)

        match = self._rx.match(path)

        if not match:
            return None

        return RouteMatch(self, match.groupdict() if self.has_params else None)


class FilterRoute(Route):
    """
    Route class that supports filters for requests.
    """

    __slots__ = (
        "handler",
        "pattern",
        "filters",
        "param_names",
        "_rx",
    )

    def __init__(
        self,
        pattern: Union[str, bytes],
        handler: Any,
        filters: List[RouteFilter],
    ):
        super().__init__(pattern, handler)
        self.filters = filters

    def match(self, request: Request) -> Optional[RouteMatch]:
        match = super().match(request)

        if not match:
            return None

        if all(action_filter.handle(request) for action_filter in self.filters):
            return match

        return None


class RouterBase(ABC):
    """
    Base abstract class for types that can register types and can implement filters.
    """

    def __init__(
        self,
        *,
        host: Optional[str] = None,
        headers: Optional[HeadersType] = None,
        params: Optional[ParamsType] = None,
        filters: Optional[List[RouteFilter]] = None,
    ):
        self._filters = normalize_filters(host, headers, params, filters)

    @abstractmethod
    def add(
        self,
        method: str,
        pattern: str,
        handler: Callable,
    ) -> None:
        """Adds a request handler for the given HTTP method and route pattern."""

    def mark_handler(self, handler: Callable) -> None:
        setattr(handler, "route_handler", True)

    def normalize_default_pattern_name(self, handler_name: str) -> str:
        return handler_name.replace("_", "-")

    def _get_decorator(
        self,
        method: str,
        pattern: Optional[str] = "/",
    ) -> Callable[..., Any]:
        def decorator(fn):
            nonlocal pattern
            if pattern is ... or pattern is None:
                # default to something depending on decorated function's name
                if fn.__name__ in {"index", "default"}:
                    pattern = "/"
                else:
                    pattern = "/" + self.normalize_default_pattern_name(fn.__name__)

                logging.debug(
                    "Defaulting to route pattern '%s' for" "request handler <%s>",
                    pattern,
                    fn.__qualname__,
                )
            self.add(method, pattern, fn)
            return fn

        return decorator

    def add_head(self, pattern: str, handler: Callable[..., Any]) -> None:
        self.add(RouteMethod.HEAD, pattern, handler)

    def add_get(self, pattern: str, handler: Callable[..., Any]) -> None:
        self.add(RouteMethod.GET, pattern, handler)

    def add_post(self, pattern: str, handler: Callable[..., Any]) -> None:
        self.add(RouteMethod.POST, pattern, handler)

    def add_put(self, pattern: str, handler: Callable[..., Any]) -> None:
        self.add(RouteMethod.PUT, pattern, handler)

    def add_delete(self, pattern: str, handler: Callable[..., Any]) -> None:
        self.add(RouteMethod.DELETE, pattern, handler)

    def add_trace(self, pattern: str, handler: Callable[..., Any]) -> None:
        self.add(RouteMethod.TRACE, pattern, handler)

    def add_options(self, pattern: str, handler: Callable[..., Any]) -> None:
        self.add(RouteMethod.OPTIONS, pattern, handler)

    def add_connect(self, pattern: str, handler: Callable[..., Any]) -> None:
        self.add(RouteMethod.CONNECT, pattern, handler)

    def add_patch(self, pattern: str, handler: Callable[..., Any]) -> None:
        self.add(RouteMethod.PATCH, pattern, handler)

    def add_ws(self, pattern: str, handler: Callable[..., Any]) -> None:
        self.add(RouteMethod.GET_WS, pattern, handler)

    def head(self, pattern: Optional[str] = "/") -> Callable[..., Any]:
        return self._get_decorator(RouteMethod.HEAD, pattern)

    def get(self, pattern: Optional[str] = "/") -> Callable[..., Any]:
        return self._get_decorator(RouteMethod.GET, pattern)

    def post(self, pattern: Optional[str] = "/") -> Callable[..., Any]:
        return self._get_decorator(RouteMethod.POST, pattern)

    def put(self, pattern: Optional[str] = "/") -> Callable[..., Any]:
        return self._get_decorator(RouteMethod.PUT, pattern)

    def delete(self, pattern: Optional[str] = "/") -> Callable[..., Any]:
        return self._get_decorator(RouteMethod.DELETE, pattern)

    def trace(self, pattern: Optional[str] = "/") -> Callable[..., Any]:
        return self._get_decorator(RouteMethod.TRACE, pattern)

    def options(self, pattern: Optional[str] = "/") -> Callable[..., Any]:
        return self._get_decorator(RouteMethod.OPTIONS, pattern)

    def connect(self, pattern: Optional[str] = "/") -> Callable[..., Any]:
        return self._get_decorator(RouteMethod.CONNECT, pattern)

    def patch(self, pattern: Optional[str] = "/") -> Callable[..., Any]:
        return self._get_decorator(RouteMethod.PATCH, pattern)

    def ws(self, pattern) -> Callable[..., Any]:
        return self._get_decorator(RouteMethod.GET_WS, pattern)

    def route(
        self, pattern: str, methods: Optional[Sequence[str]] = None
    ) -> Callable[..., Any]:
        if methods is None:
            methods = ["GET"]

        def decorator(f):
            for method in methods:
                self.add(method, pattern, f)
            return f

        return decorator


class MultiRouterMixin:
    """
    This mixin is activate automatically when a Router defines sub-routers.
    """

    _sub_routers: List["Router"]

    def __iter__(self):
        yield from super().__iter__()  # type: ignore

        for router in self._sub_routers:
            yield from router

    def iter_with_methods(self):
        """
        Iters through the routes defined in this Router, yielding each route
        and its HTTP method.
        """
        yield from super().__iter__()  # type: ignore

        for router in self._sub_routers:
            yield from router.iter_with_methods()

    def get_match(self, request: Request) -> Optional[RouteMatch]:
        for router in self._sub_routers:
            match = router.get_match(request)

            if match:
                return match

        return super().get_match(request)  # type: ignore


class RouterFiltersMixin:
    """
    This mixin is activated automatically when any of the routes defined for a web app
    uses filters (RouteFilter). The rationale for using a mixin is that using filters
    incurs a performance fee, and fees should only be paid when using features.
    """

    routes: Dict[bytes, List[Route]]
    _fallback: Any

    def get_match(self, request: Request) -> Optional[RouteMatch]:
        for route in self.routes[ensure_bytes(request.method)]:
            match = route.match(request)

            if match:
                return match

        if self._fallback is None:
            return None

        return RouteMatch(self._fallback, None)


RouteConfig = Union[Dict[str, Any], "Router"]


class Router(RouterBase):
    __slots__ = (
        "routes",
        "controllers_routes",
        "_map",
        "_fallback",
        "_sub_routers",
        "_filters",
        "_path_to_methods",
    )

    def __init__(
        self,
        *,
        host: Optional[str] = None,
        headers: Optional[HeadersType] = None,
        params: Optional[ParamsType] = None,
        filters: Optional[List[RouteFilter]] = None,
        sub_routers: Optional[List["Router"]] = None,
    ):
        super().__init__(
            host=host,
            headers=headers,
            params=params,
            filters=filters,
        )

        self._map = {}
        self._fallback = None
        self.routes: Dict[bytes, List[Route]] = defaultdict(list)
        self.controllers_routes = RoutesRegistry()
        self._sub_routers = sub_routers
        self._path_to_methods: Dict[bytes, Set[str]] = defaultdict(set)

        if self._filters:
            extend(self, RouterFiltersMixin)

        if self._sub_routers:
            extend(self, MultiRouterMixin)

    def reset(self):
        """Resets this router to its initial state."""
        self._map = {}
        self._fallback = None
        self.routes = defaultdict(list)
        self.controllers_routes.reset()
        if self._sub_routers:
            for sub_router in self._sub_routers:
                sub_router.reset()

    @property
    def fallback(self):
        return self._fallback

    @fallback.setter
    def fallback(self, value):
        if not isinstance(value, Route):
            if callable(value):
                self._fallback = Route(b"*", value)
                return
            raise ValueError("fallback must be a Route or a callable")
        self._fallback = value

    def __iter__(self):
        for _, routes in self.routes.items():
            for route in routes:
                yield route
        if self._fallback:
            yield self._fallback

    def iter_with_methods(self):
        """
        Iters through the routes defined in this Router, yielding each route
        and its HTTP method.
        """
        for method, routes in self.routes.items():
            for route in routes:
                yield method.decode(), route
        if self._fallback:
            yield "*", self._fallback

    def _is_route_configured(self, method: bytes, route: Route):
        if isinstance(route, FilterRoute):  # pragma: no cover
            # The route defines action filters, we cannot determine if the user is
            # registering twice the same pattern. However, the user opted-in for an
            # advanced feature and should be aware about conflicting routes.
            return False

        method_patterns = self._map.get(method)
        if not method_patterns:
            return False
        existing_route: Route = method_patterns.get(route.full_pattern)

        if existing_route and not isinstance(existing_route, FilterRoute):
            return True
        return False

    def _set_configured_route(self, method: bytes, route: Route):
        method_patterns = self._map.get(method)
        if not method_patterns:
            self._map[method] = {route.full_pattern: route}
        else:
            method_patterns[route.full_pattern] = route

    def _check_duplicate(self, method: bytes, new_route: Route):
        if self._is_route_configured(method, new_route):
            current_route = self._map[method][new_route.full_pattern]
            raise RouteDuplicate(method, new_route.pattern, current_route.handler)
        self._set_configured_route(method, new_route)

    def add(
        self,
        method: str,
        pattern: AnyStr,
        handler: Any,
        filters: Optional[List[RouteFilter]] = None,
    ):
        if filters and not isinstance(self, RouterFiltersMixin):
            extend(self, RouterFiltersMixin)

        route_filters = filters or self._filters

        self.mark_handler(handler)
        method_name = ensure_bytes(method)
        new_route = (
            FilterRoute(ensure_bytes(pattern), handler, route_filters)
            if route_filters
            else Route(ensure_bytes(pattern), handler)
        )

        if not route_filters:
            self._check_duplicate(method_name, new_route)

        self.add_route(method_name, new_route)

        # Update the set of allowed methods for a given path
        self._path_to_methods[new_route.pattern].add(method)

    def add_route(self, method: AnyStr, route: Route):
        self.routes[ensure_bytes(method)].append(route)

    def sort_routes(self):
        """
        Sorts the current routes in order of dynamic parameters ascending.
        The catch-all route is always placed to the end.
        """
        current_routes = self.routes.copy()

        for method in current_routes.keys():
            current_routes[method].sort(key=lambda route: -route.pattern.count(b"/"))
            current_routes[method].sort(key=lambda route: len(route.param_names))
            current_routes[method].sort(key=lambda route: b".*" in route.rx.pattern)
            current_routes[method].sort(
                key=lambda route: (
                    -len(route.filters) if isinstance(route, FilterRoute) else 0
                )
            )
            current_routes[method].sort(
                key=lambda route: b"*" == route.pattern or b"/*" == route.pattern
            )

        self.routes = current_routes

        if self._sub_routers:
            for sub_router in self._sub_routers:
                sub_router.sort_routes()
    
    @lru_cache(maxsize=100)
    def get_allowed_methods(self, path: bytes) -> Set[str]:
        """
        Returns a set of allowed methods for a given path.
        """
        return self._path_to_methods.get(path, set())

    def get_match(self, request: Request) -> Optional[RouteMatch]:
        """
        Gets a match for the given request, by method and request path.
        """
        match = self.get_match_by_method_and_path(request.method, request._path)
        if not match:
            # Check if the path exists but with different methods
            allowed_methods = self.get_allowed_methods(request._path)
            if allowed_methods:
                request.allowed_methods = allowed_methods  # Attach allowed methods to request for later use
        return match

    # def get_match(self, request: Request) -> Optional[RouteMatch]:
    #     """
    #     Gets a match for the given request, by method and request path.
    #     """
    #     return self.get_match_by_method_and_path(request.method, request._path)

    @lru_cache(maxsize=1200)
    def get_match_by_method_and_path(
        self, method: AnyStr, path: AnyStr
    ) -> Optional[RouteMatch]:
        bytes_value = ensure_bytes(path)
        for route in self.routes[ensure_bytes(method)]:
            match = route.match_by_path(bytes_value)
            if match:
                return match
        if self._fallback is None:
            return None

        return RouteMatch(self._fallback, None)

    @lru_cache(maxsize=1200)
    def get_matching_route(self, method: AnyStr, value: AnyStr) -> Optional[Route]:
        for route in self.routes[ensure_bytes(method)]:
            match = route.match_by_path(ensure_bytes(value))
            if match:
                return route
        return None


class RegisteredRoute:
    __slots__ = ("method", "pattern", "handler")

    def __init__(
        self,
        method: str,
        pattern: str,
        handler: Callable,
    ):
        self.method = method
        self.pattern = pattern
        self.handler = handler


class RoutesRegistry(RouterBase):
    """
    A registry for routes: not a full router able to get matches.
    Unlike a router, a registry does not throw for duplicated routes;
    because such routes can be modified when applied to an actual router.

    This class is meant to enable scenarios like base pattern for controllers.
    """

    __slots__ = ("routes", "_filters")

    def __init__(
        self,
        *,
        host: Optional[str] = None,
        headers: Optional[HeadersType] = None,
        params: Optional[ParamsType] = None,
        filters: Optional[List[RouteFilter]] = None,
    ):
        super().__init__(host=host, headers=headers, params=params, filters=filters)
        self.routes: List[RegisteredRoute] = []

    def reset(self):
        """Resets this routes registry to its initial state."""
        self.routes = []

    def __iter__(self):
        yield from self.routes

    def add(
        self,
        method: str,
        pattern: str,
        handler: Callable,
    ):
        self.mark_handler(handler)
        self.routes.append(RegisteredRoute(method, pattern, handler))


class MountRegistry:
    """
    Holds information about mounted applications and how they should be handled.
    """

    __slots__ = ("_mounted_apps", "_mounted_paths", "auto_events", "handle_docs")

    def __init__(self, auto_events: bool = True, handle_docs: bool = False):
        self._mounted_apps = []
        self._mounted_paths = set()
        self.auto_events = auto_events
        self.handle_docs = handle_docs

    @property
    def mounted_apps(self) -> List[Route]:
        return self._mounted_apps

    @property
    def mounted_paths(self) -> Set[str]:
        return self._mounted_paths

    def mount(self, path: str, app: Callable) -> None:
        if not path:
            path = "/"

        if path.lower() in self._mounted_paths:
            raise AssertionError(f"Mount application with path '{path}' already exist")

        self._mounted_paths.add(path.lower())

        if not path.endswith("/*"):
            path = f"{path.rstrip('/*')}/*"

        self._mounted_apps.append(Route(path, app))


# For backward compatibility
Mount = MountRegistry


_apps_by_router_id = {}


def validate_router(app):
    """
    Ensures that the same router is not bound to more than one application object.
    If the same application is being reloaded dynamically, like when using uvicorn
    programmatically, the router is reset.
    """
    app_router: Router = app.router
    router_id = id(app_router)

    # Get information about where the application was instantiated (which filename,
    # which line_number)
    _, filename, line_number, *_ = inspect.stack()[2]

    try:
        existing_app = _apps_by_router_id[router_id]
    except KeyError:
        # Good
        _apps_by_router_id[router_id] = {
            "app": app,
            "filename": filename,
            "line_number": line_number,
        }
    else:
        if (
            existing_app["filename"] == filename
            and existing_app["line_number"] == line_number
        ):
            # This is the same application! This can happen when imported dynamically
            # by uvicorn reload feature, when uvicorn is started programmatically.
            # See https://github.com/Neoteroi/BlackSheep/issues/438
            logging.getLogger("shuttleasgi.server").warning(
                "The application was reloaded, resetting its router."
            )
            app_router.reset()
            return
        raise SharedRouterError()


def validate_default_router():
    """
    This method ensures that the default router is associated to an application, if it
    defines any route.
    """
    if set(router):
        # The default router has routes defined, ensure that it is bound to an
        # application
        # verify that
        try:
            _apps_by_router_id[id(router)]
        except KeyError:
            # Not good
            raise OrphanDefaultRouterError() from None


# Singleton router used to store initial configuration,
# before the application starts
# this is used as *default* router, but it can be overridden;
# This is done for two reasons: to reduce code verbosity when defining routes,
# and because we can expect that in most use cases web applications use a single
# Application and Router (the same approach can be easily used for more complex use
# cases where more than one router is used).
router = Router()
controllers_routes = router.controllers_routes

head = router.head
get = router.get
post = router.post
put = router.put
patch = router.patch
delete = router.delete
trace = router.trace
options = router.options
connect = router.connect
ws = router.ws
route = router.route



"""
Provides functions related to the server process.
"""

import os
import signal
from typing import TYPE_CHECKING

from shuttleasgi.utils import truthy

if TYPE_CHECKING:
    from shuttleasgi.server.application import Application

_STOPPING = False


def is_stopping() -> bool:
    """
    Returns a value indicating whether the server process received a SIGINT or a SIGTERM
    signal, and therefore the application is stopping.
    """
    if not truthy(os.environ.get("APP_SIGNAL_HANDLER", "")):
        raise RuntimeError(
            "This function can only be used if the env variable `APP_SIGNAL_HANDLER=1`"
        )
    return _STOPPING


def use_shutdown_handler(app: "Application"):
    """
    Configures an application start event handler that listens to SIGTERM and SIGINT
    to know when the process is stopping.
    """

    @app.on_start
    async def configure_shutdown_handler():
        # See the conversation here:
        # https://github.com/encode/uvicorn/issues/1579#issuecomment-1419635974
        for signal_type in {signal.SIGINT, signal.SIGTERM}:
            current_handler = signal.getsignal(signal_type)

            def terminate_now(signum, frame):
                global _STOPPING
                _STOPPING = True

                if callable(current_handler):
                    current_handler(signum, frame)  # type: ignore

            signal.signal(signal_type, terminate_now)



from abc import ABC, abstractmethod
from typing import Any


class ModelHandler(ABC):
    """Type that can handle a model object for a view template."""

    @abstractmethod
    def model_to_view_params(self, model: Any) -> Any:
        """Converts a model to parameters for a template view."""


class Renderer(ABC):
    """Type that can render HTML views."""

    @abstractmethod
    def render(self, template: str, model, **kwargs) -> str:
        """Renders a view synchronously."""

    @abstractmethod
    async def render_async(self, template: str, model, **kwargs) -> str:
        """Renders a view asynchronously."""

    @abstractmethod
    def bind_antiforgery_handler(self, handler) -> None:
        """Applies extensions for an antiforgery handler."""



"""
This module offer built-in functions for Server Sent Events.
"""

from typing import AsyncIterable, Callable, List, Optional, Tuple

from shuttleasgi.contents import ServerSentEvent, StreamedContent
from shuttleasgi.messages import Response
from shuttleasgi.scribe import write_sse

__all__ = [
    "ServerSentEvent",
    "ServerSentEventsContent",
    "ServerSentEventsResponse",
    "EventsProvider",
]


EventsProvider = Callable[[], AsyncIterable[ServerSentEvent]]


class ServerSentEventsContent(StreamedContent):
    """
    A specialized kind of StreamedContent that can be used to stream
    Server-Sent Events to a client.
    """

    def __init__(self, events_provider: EventsProvider):
        super().__init__(b"text/event-stream", self.write_events(events_provider))

    @staticmethod
    def write_events(
        events_provider: EventsProvider,
    ) -> Callable[[], AsyncIterable[bytes]]:
        async def write_events():
            async for event in events_provider():
                yield write_sse(event)

        return write_events


class ServerSentEventsResponse(Response):
    """
    An Response type that can be used to stream Server-Sent Events to a client.
    """

    def __init__(
        self,
        events_provider: EventsProvider,
        status: int = 200,
        headers: Optional[List[Tuple[bytes, bytes]]] = None,
    ) -> None:
        if headers is None:
            headers = [
                (b"Cache-Control", b"no-cache"),
                (b"Connection", b"Keep-Alive"),
                (b"X-Accel-Buffering", b"no")
            ]
        super().__init__(status, headers, ServerSentEventsContent(events_provider))



import orjson
from typing import Any

from essentials.json import dumps


def default_json_dumps(obj) -> str:
    return orjson.dumps(
        obj,
        option=orjson.OPT_SERIALIZE_NUMPY | orjson.OPT_SERIALIZE_UUID
    ).decode("utf-8")


def default_pretty_json_dumps(obj) -> str:
    return orjson.dumps(
        obj,
        option=orjson.OPT_SERIALIZE_NUMPY
            | orjson.OPT_SERIALIZE_UUID
            | orjson.OPT_INDENT_2
    ).decode("utf-8")


def orjson_dumps(obj) -> bytes:
    return orjson.dumps(
        obj,
        option=orjson.OPT_SERIALIZE_NUMPY | orjson.OPT_SERIALIZE_UUID
    )


def orjson_pretty_dumps(obj) -> bytes:
    return orjson.dumps(
        obj,
        option=orjson.OPT_SERIALIZE_NUMPY
            | orjson.OPT_SERIALIZE_UUID
            | orjson.OPT_INDENT_2
    )


def builtin_json_dumps(obj) -> str:
    return dumps(obj, separators=(",", ":"))


def builtin_pretty_json_dumps(obj) -> str:
    return dumps(obj, indent=4)


class JSONSettings:
    def __init__(self):
        self._loads = orjson.loads
        self._dumps = default_json_dumps
        self._pretty_dumps = default_pretty_json_dumps
        self._odumps = orjson_dumps
        self._opretty_dumps = orjson_pretty_dumps

    def use(
        self,
        loads=orjson.loads,
        dumps=default_json_dumps,
        pretty_dumps=default_pretty_json_dumps,
    ):
        self._loads = loads
        self._dumps = dumps
        self._pretty_dumps = pretty_dumps
        self._odumps = orjson_dumps
        self._opretty_dumps = orjson_pretty_dumps

    def loads(self, text: str) -> Any:
        return self._loads(text)

    def dumps(self, obj: Any) -> str:
        return self._dumps(obj)

    def pretty_dumps(self, obj: Any) -> str:
        return self._pretty_dumps(obj)

    def odumps(self, obj: Any) -> bytes:
        return self._odumps(obj)

    def opretty_dumps(self, obj: Any) -> bytes:
        return self._opretty_dumps(obj)


json_settings = JSONSettings()



from dataclasses import asdict, is_dataclass
from typing import Any

from .abc import ModelHandler


class DefaultModelHandler(ModelHandler):
    def model_to_view_params(self, model: Any) -> Any:
        if isinstance(model, dict):
            return model
        if is_dataclass(model):
            return asdict(model)
        if hasattr(model, "__dict__"):
            return model.__dict__
        return model



from typing import Callable, TypeVar

from rodi import Container, ContainerProtocol

T = TypeVar("T")


def default_container_factory() -> ContainerProtocol:
    return Container()


class DISettings:
    def __init__(self):
        self._container_factory = default_container_factory

    def use(self, container_factory: Callable[[], ContainerProtocol]):
        self._container_factory = container_factory

    def get_default_container(self) -> ContainerProtocol:
        return self._container_factory()


di_settings = DISettings()



from shuttleasgi.contents import FormContent, JSONContent, TextContent
from shuttleasgi.testing.client import TestClient
from shuttleasgi.testing.messages import MockReceive, MockSend
from shuttleasgi.testing.simulator import AbstractTestSimulator

__all__ = [
    "TestClient",
    "AbstractTestSimulator",
    "JSONContent",
    "TextContent",
    "FormContent",
    "MockReceive",
    "MockSend",
]



from setuptools import Extension, setup

COMPILE_ARGS = ["-O2"]


setup(
    ext_modules=[
        Extension(
            "shuttleasgi.url",
            ["shuttleasgi/url.c"],
            extra_compile_args=COMPILE_ARGS,
        ),
        Extension(
            "shuttleasgi.exceptions",
            ["shuttleasgi/exceptions.c"],
            extra_compile_args=COMPILE_ARGS,
        ),
        Extension(
            "shuttleasgi.headers",
            ["shuttleasgi/headers.c"],
            extra_compile_args=COMPILE_ARGS,
        ),
        Extension(
            "shuttleasgi.cookies",
            ["shuttleasgi/cookies.c"],
            extra_compile_args=COMPILE_ARGS,
        ),
        Extension(
            "shuttleasgi.contents",
            ["shuttleasgi/contents.c"],
            extra_compile_args=COMPILE_ARGS,
        ),
        Extension(
            "shuttleasgi.messages",
            ["shuttleasgi/messages.c"],
            extra_compile_args=COMPILE_ARGS,
        ),
        Extension(
            "shuttleasgi.scribe",
            ["shuttleasgi/scribe.c"],
            extra_compile_args=COMPILE_ARGS,
        ),
        Extension(
            "shuttleasgi.baseapp",
            ["shuttleasgi/baseapp.c"],
            extra_compile_args=COMPILE_ARGS,
        ),
    ]
)



from typing import TYPE_CHECKING, Awaitable, Callable

from rodi import ActivationScope, Container

from shuttleasgi.messages import Request, Response

if TYPE_CHECKING:
    from shuttleasgi.server.application import Application


async def di_scope_middleware(
    request: Request, handler: Callable[[Request], Awaitable[Response]]
) -> Response:
    """
    This middleware ensures that a single scope is used for Dependency Injection,
    across request handlers and other parts of the application that require activating
    services (e.g. authentication handlers).

    This middleware is not necessary in most cases, but in some circumstances it can be
    necessary.
    """
    with ActivationScope() as scope:
        scope.scoped_services[Request] = request  # type: ignore
        scope.scoped_services["__request__"] = request  # type: ignore
        request._di_scope = scope  # type: ignore
        return await handler(request)


def request_factory(context) -> Request:
    # The following scoped service is set in a middleware, since in fact we are
    # mixing runtime data with composition data.
    return context.scoped_services[Request]


def register_http_context(app: "Application"):
    """
    Makes the `Request` object accessible through dependency injection for the
    application container.
    This method requires using `rodi` as solution for dependency injection, since
    other implementations might not support scoped services and factories using the
    activation scope.

    This is not a recommended pattern, but it might be desired in certain situations.
    """
    assert isinstance(app.services, Container), "This method requires rodi."

    if di_scope_middleware not in app.middlewares:

        @app.on_middlewares_configuration
        def enable_request_accessor(_):
            app.middlewares.insert(0, di_scope_middleware)

    app.services.add_scoped_by_factory(request_factory)



import os
from functools import lru_cache
from typing import Optional

from jinja2 import Environment, PackageLoader, Template, nodes, select_autoescape
from jinja2.ext import Extension
from jinja2.utils import pass_context

from shuttleasgi.messages import Request
from shuttleasgi.server.csrf import AntiForgeryHandler, MissingRequestContextError

from .abc import Renderer

_DEFAULT_TEMPLATES_EXTENSION = os.environ.get("APP_JINJA_EXTENSION", ".jinja")


@lru_cache(1200)
def get_template_name(name: str) -> str:
    if not name.endswith(_DEFAULT_TEMPLATES_EXTENSION):
        return name + _DEFAULT_TEMPLATES_EXTENSION
    return name


def render_template(template: Template, *args, **kwargs):
    return template.render(*args, **kwargs)


async def render_template_async(template: Template, *args, **kwargs):
    return await template.render_async(*args, **kwargs)


class AntiForgeryBaseExtension(Extension):
    af_handler: Optional[AntiForgeryHandler] = None

    def __init__(self, environment):
        super().__init__(environment)

        if self.af_handler is None:
            raise TypeError("Define a subclass bound to an AntiForgeryHandler")

    @property
    def handler(self) -> AntiForgeryHandler:
        if self.af_handler is None:
            raise TypeError("Missing anti_forgery_handler")
        return self.af_handler

    def parse(self, parser):
        line_number = next(parser.stream).lineno
        return nodes.CallBlock(self.call_method("get_html"), [], [], "").set_lineno(
            line_number
        )

    def get_token(self, context) -> str:
        try:
            request = context["request"]
        except KeyError:
            raise MissingRequestContextError()
        assert isinstance(request, Request)

        tokens = self.handler.get_tokens(request)
        return tokens[1]


class AntiForgeryInputExtension(AntiForgeryBaseExtension):
    tags = {"csrf_input", "af_input"}

    @pass_context
    def get_html(self, context, caller):
        value = self.get_token(context)
        return (
            f'<input type="hidden" name="{self.handler.form_name}" '
            f'value="{value}" />'
        )


class AntiForgeryValueExtension(AntiForgeryBaseExtension):
    tags = {"csrf_token", "af_token"}

    @pass_context
    def get_html(self, context, caller):
        return self.get_token(context)


class JinjaRenderer(Renderer):
    def __init__(
        self,
        loader: Optional[PackageLoader] = None,
        debug: bool = False,
        enable_async: bool = False,
    ) -> None:
        super().__init__()
        self.env = Environment(
            loader=loader
            or PackageLoader(
                os.environ.get("APP_JINJA_PACKAGE_NAME", "app"),
                os.environ.get("APP_JINJA_PACKAGE_PATH", "views"),
            ),
            autoescape=select_autoescape(["html", "xml", "jinja"]),
            auto_reload=bool(os.environ.get("APP_JINJA_DEBUG", "")) or debug,
            enable_async=bool(os.environ.get("APP_JINJA_ENABLE_ASYNC", ""))
            or enable_async,
        )

    def render(self, template: str, model, **kwargs) -> str:
        if model:
            return render_template(
                self.env.get_template(get_template_name(template)), model, **kwargs
            )
        return render_template(
            self.env.get_template(get_template_name(template)), **kwargs
        )

    async def render_async(self, template: str, model, **kwargs) -> str:
        if model:
            return await self.env.get_template(
                get_template_name(template)
            ).render_async(model, **kwargs)
        return await self.env.get_template(get_template_name(template)).render_async(
            **kwargs
        )

    def bind_antiforgery_handler(self, handler) -> None:
        class BoundAntiForgeryInputExtension(AntiForgeryInputExtension):
            af_handler = handler

        class BoundAntiForgeryValueExtension(AntiForgeryValueExtension):
            af_handler = handler

        self.env.add_extension(BoundAntiForgeryInputExtension)
        self.env.add_extension(BoundAntiForgeryValueExtension)
